from typing import Optional


class PrefixLengthError(Exception):
    pass


class PrefixKeyError(Exception):
    pass


class Prefix:

    """A class to generate an identifier prefix."""

    template: str = "{protocol_number}{requisition_identifier}"
    length: int = 10

    def __init__(
        self, template: Optional[str] = None, length: Optional[int] = None, **template_opts
    ):
        self.template: str = template or self.template
        self.length: int = length or self.length
        template_opts = {k: v for k, v in template_opts.items() if v is not None}
        try:
            self.prefix: str = self.template.format(**template_opts)
        except KeyError as e:
            raise PrefixKeyError(
                f"Missing template value for '{e}'. Got options={template_opts}"
            )
        if len(self.prefix) != self.length:
            raise PrefixLengthError(
                f"Invalid prefix '{self.prefix}'. "
                f"Got length == {len(self.prefix)}. Expected {self.length}."
            )

    def __repr__(self):
        return f"{self.__class__.__name__}({self.template}, {self.length})"

    def __str__(self):
        return self.prefix



from edc_lab import RequisitionPanelGroup
from edc_lab.lab import AliquotType, Process, ProcessingProfile, RequisitionPanel


class EdcLabTestMixin:
    @staticmethod
    def get_panel_group():
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        b = AliquotType(name="aliquot_b", numeric_code="66", alpha_code="BB")
        a.add_derivatives(b)
        process = Process(aliquot_type=b, aliquot_count=3)
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        processing_profile.add_processes(process)
        rft_panel = RequisitionPanel(
            name="chemistry_rft",
            verbose_name="Chemistry: Renal Function Tests",
            abbreviation="RFT",
            processing_profile=processing_profile,
            utest_ids=["urea", "creatinine", "uric_acid", "egfr"],
        )

        lipids_panel = RequisitionPanel(
            name="chemistry_lipids",
            verbose_name="Chemistry: Lipids",
            abbreviation="LIPIDS",
            processing_profile=processing_profile,
            utest_ids=["ldl", "hdl", "trig"],
        )

        lft_panel = RequisitionPanel(
            name="chemistry_lft",
            verbose_name="Chemistry: Liver Function Tests",
            abbreviation="LFT",
            processing_profile=processing_profile,
            utest_ids=["ast", "alt", "alp", "amylase", "ggt", "albumin"],
        )

        return RequisitionPanelGroup(
            lft_panel,
            rft_panel,
            lipids_panel,
            name="chemistry",
            verbose_name="Chemistry: LFT, RFT, Lipids",
            reference_range_collection_name="default",
        )



from django.conf import settings
from django.contrib.auth.models import User
from django.test import TestCase
from django.test.utils import override_settings
from edc_sites import add_or_update_django_sites
from edc_sites.single_site import SingleSite
from edc_sites.tests import SiteTestCaseMixin
from multisite import SiteID

from edc_lab.models import (
    Aliquot,
    Box,
    BoxItem,
    BoxType,
    Consignee,
    Manifest,
    ManifestItem,
    Shipper,
)
from edc_lab.reports import ManifestReport, ManifestReportError


class TestManifest(SiteTestCaseMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        add_or_update_django_sites(
            sites=[
                SingleSite(
                    settings.SITE_ID,
                    "test_site",
                    country_code="ug",
                    country="uganda",
                    domain="bugamba.ug.clinicedc.org",
                )
            ]
        )
        return super().setUpClass()

    def test_manifest(self):
        consignee = Consignee.objects.create(name="consignee")
        shipper = Shipper.objects.create(name="shipper")
        Manifest.objects.create(consignee=consignee, shipper=shipper)

    def test_manifest_with_items(self):
        consignee = Consignee.objects.create(name="consignee")
        shipper = Shipper.objects.create(name="shipper")
        manifest = Manifest.objects.create(consignee=consignee, shipper=shipper)
        ManifestItem.objects.create(manifest=manifest, identifier="aaaaaaaaaaaa")

    def test_manifest_with_items_slug(self):
        consignee = Consignee.objects.create(name="consignee")
        shipper = Shipper.objects.create(name="shipper")
        manifest = Manifest.objects.create(consignee=consignee, shipper=shipper)
        manifest_item = ManifestItem.objects.create(
            manifest=manifest, identifier="aaaaaaaaaaaabb"
        )
        self.assertIn("aaaaaaaaaaaabb", manifest_item.slug)
        self.assertIn(manifest_item.human_readable_identifier, manifest_item.slug)


class TestManifestReport(SiteTestCaseMixin, TestCase):
    def setUp(self):
        add_or_update_django_sites(sites=self.default_sites, verbose=False)
        self.user = User.objects.create(first_name="Noam", last_name="Chomsky")
        consignee = Consignee.objects.create(name="consignee")
        shipper = Shipper.objects.create(name="shipper")
        self.manifest = Manifest.objects.create(consignee=consignee, shipper=shipper)

    @override_settings(SITE_ID=SiteID(default=20))
    def test_report(self):
        self.assertEqual(self.manifest.site.name, "test_site")
        report = ManifestReport(manifest=self.manifest, user=self.user)
        report.render()

    @override_settings(SITE_ID=SiteID(default=20))
    def test_report_shipped(self):
        self.manifest.shipped = True
        self.manifest.save()
        self.assertEqual(self.manifest.site.name, "test_site")
        report = ManifestReport(manifest=self.manifest, user=self.user)
        report.render()

    @override_settings(SITE_ID=SiteID(default=20))
    def test_report_items_not_in_box(self):
        self.manifest.shipped = True
        self.manifest.save()
        for i in range(0, 3):
            ManifestItem.objects.create(
                manifest=self.manifest,
                identifier=f"{self.manifest.manifest_identifier}{i}",
            )
        self.assertEqual(self.manifest.site.name, "test_site")
        report = ManifestReport(manifest=self.manifest, user=self.user)
        self.assertRaises(ManifestReportError, report.render)
        try:
            report.render()
        except ManifestReportError as e:
            self.assertEqual(e.code, "unboxed_item")

    def test_box_type(self):
        BoxType.objects.create(name="box_type", across=8, down=8, total=64)

    def test_box(self):
        box_type = BoxType.objects.create(name="box_type", across=8, down=8, total=64)
        Box.objects.create(box_type=box_type)

    def test_box_item(self):
        box_type = BoxType.objects.create(name="box_type", across=8, down=8, total=64)
        box = Box.objects.create(box_type=box_type)
        BoxItem.objects.create(box=box, identifier=box.box_identifier, position=0)

    @override_settings(SITE_ID=SiteID(default=20))
    def test_report_invalid_invalid_aliquot_identifier(self):
        self.manifest.shipped = True
        self.manifest.save()
        box_type = BoxType.objects.create(name="box_type", across=8, down=8, total=64)
        box = Box.objects.create(box_type=box_type)
        # add box items with invalid aliquot identifiers
        for i in range(0, 3):
            BoxItem.objects.create(box=box, identifier=f"{i}", position=i)
        ManifestItem.objects.create(manifest=self.manifest, identifier=box.box_identifier)
        self.assertEqual(self.manifest.site.name, "test_site")
        report = ManifestReport(manifest=self.manifest, user=self.user)
        self.assertRaises(ManifestReportError, report.render)
        try:
            report.render()
        except ManifestReportError as e:
            self.assertEqual(e.code, "invalid_aliquot_identifier")

    @override_settings(SITE_ID=SiteID(default=20))
    def test_report_invalid_invalid_requisition_identifier(self):
        self.manifest.shipped = True
        self.manifest.save()
        prefix = "ABCDEFG"
        for i in range(0, 3):
            Aliquot.objects.create(count=i, aliquot_identifier=f"{prefix}{i}")
        box_type = BoxType.objects.create(name="box_type", across=8, down=8, total=64)
        box = Box.objects.create(box_type=box_type)
        # add box items with invalid aliquot identifiers
        for index, aliquot in enumerate(Aliquot.objects.all()):
            BoxItem.objects.create(
                box=box, identifier=aliquot.aliquot_identifier, position=index
            )
        ManifestItem.objects.create(manifest=self.manifest, identifier=box.box_identifier)
        self.assertEqual(self.manifest.site.name, "test_site")
        report = ManifestReport(manifest=self.manifest, user=self.user)
        self.assertRaises(ManifestReportError, report.render)
        try:
            report.render()
        except ManifestReportError as e:
            self.assertEqual(e.code, "invalid_requisition_identifier")



from edc_constants.constants import COMPLETE, NOT_APPLICABLE, OTHER, PARTIAL, PENDING
from edc_metadata.constants import NOT_REQUIRED
from edc_reportable import (
    MICROMOLES_PER_LITER,
    MICROMOLES_PER_LITER_DISPLAY,
    MILLIGRAMS_PER_DECILITER,
    MILLIMOLES_PER_LITER,
    MILLIMOLES_PER_LITER_DISPLAY,
)

from .constants import EQ, FILL_ACROSS, FILL_DOWN, FINGER_PRICK, GT, GTE, LT, LTE, TUBE

ABS_CALC = (("absolute", "Absolute"), ("calculated", "Calculated"))

ALIQUOT_STATUS = (("available", "available"), ("consumed", "consumed"))

ALIQUOT_CONDITIONS = (
    ("10", "OK"),
    ("20", "Inadequate volume for testing"),
    ("30", "Clotted or haemolised"),
    ("40", "Wrong tube type, unable to test"),
    ("50", "Sample degradation has occured. Unsuitable for testing"),
    ("60", "Expired tube"),
    ("70", "Technical problem at lab, unable to test"),
)

FILL_ORDER = ((FILL_ACROSS, "Across"), (FILL_DOWN, "Down"))

MODIFY_ACTIONS = (
    ("INSERT", "Insert"),
    ("UPDATE", "Update"),
    ("DELETE", "Delete"),
    ("PRINT", "Print"),
    ("VIEW", "Print"),
)

ORDER_STATUS = ((PENDING, "Pending"), (PARTIAL, "Partial"), (COMPLETE, "Complete"))

GLUCOSE_UNITS = (
    (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
    (MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER_DISPLAY),
)

GLUCOSE_UNITS_NA = (
    (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
    (MILLIMOLES_PER_LITER, MILLIMOLES_PER_LITER_DISPLAY),
    (NOT_APPLICABLE, "Not applicable"),
)
SERUM_CREATININE_UNITS = (
    (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
    (MICROMOLES_PER_LITER, MICROMOLES_PER_LITER_DISPLAY),
)

SERUM_CREATININE_UNITS_NA = (
    (MILLIGRAMS_PER_DECILITER, MILLIGRAMS_PER_DECILITER),
    (MICROMOLES_PER_LITER, MICROMOLES_PER_LITER_DISPLAY),
    (NOT_APPLICABLE, "Not applicable"),
)


REASON_NOT_DRAWN = (
    (NOT_APPLICABLE, "Not applicable"),
    ("collection_failed", "Tried, but unable to obtain sample from patient"),
    ("absent", "Patient did not attend visit"),
    ("refused", "Patient refused"),
    ("no_supplies", "No supplies"),
    (NOT_REQUIRED, "No longer required for this visit"),
    (OTHER, "Other"),
)

RESULT_RELEASE_STATUS = (
    ("NEW", "New"),
    ("RELEASED", "Released"),
    ("AMENDED", "Amended"),
)

RESULT_VALIDATION_STATUS = (("P", "Preliminary"), ("F", "Final"), ("R", "Rejected"))

RESULT_QUANTIFIER = ((EQ, EQ), (GT, GT), (GTE, GTE), (LT, LT), (LTE, LTE))

RESULT_QUANTIFIER_NA = (
    (NOT_APPLICABLE, ""),
    (EQ, EQ),
    (GT, GT),
    (GTE, GTE),
    (LT, LT),
    (LTE, LTE),
)

VL_QUANTIFIER_NA = (
    (EQ, EQ),
    (GT, GT),
    (LT, LT),
)


SPECIMEN_MEASURE_UNITS = (
    ("mL", "mL"),
    ("uL", "uL"),
    ("spots", "spots"),
    ("n/a", "Not Applicable"),
)

SPECIMEN_MEDIUM = (
    ("tube_any", "Tube"),
    ("tube_edta", "Tube EDTA"),
    ("swab", "Swab"),
    ("dbs_card", "DBS Card"),
)

UNITS = (
    ("%", "%"),
    ("10^3/uL", "10^3/uL"),
    ("10^3uL", "10^3uL"),
    ("10^6/uL", "10^6/uL"),
    ("cells/ul", "cells/ul"),
    ("copies/ml", "copies/ml"),
    ("fL", "fL"),
    ("g/dL", "g/dL"),
    ("g/L", "g/L"),
    ("mg/L", "mg/L"),
    ("mm/H", "mm/H"),
    ("mmol/L", "mmol/L"),
    ("ng/ml", "ng/ml"),
    ("pg", "pg"),
    ("ratio", "ratio"),
    ("U/L", "U/L"),
    ("umol/L", "umol/L"),
)

PRIORITY = (("normal", "Normal"), ("urgent", "Urgent"))

REASON_NOT_DRAWN = (
    (NOT_APPLICABLE, "Not applicable"),
    ("collection_failed", "Tried, but unable to obtain sample from patient"),
    ("absent", "Patient did not attend visit"),
    ("refused", "Patient refused"),
    ("no_supplies", "No supplies"),
    (NOT_REQUIRED, "No longer required for this visit"),
    (OTHER, "Other"),
)

ITEM_TYPE = (
    (NOT_APPLICABLE, "Not applicable"),
    (TUBE, "Tube"),
    (FINGER_PRICK, "Finger prick"),
    ("swab", "Swab"),
    ("dbs", "DBS Card"),
    (OTHER, "Other"),
)



# Generated by Django 2.0.1 on 2018-01-23 19:36

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("edc_lab", "0013_auto_20180117_1438")]

    operations = [
        migrations.RemoveField(model_name="historicalmanifest", name="site_code"),
        migrations.RemoveField(model_name="historicalmanifest", name="site_name"),
        migrations.RemoveField(model_name="manifest", name="site_code"),
        migrations.RemoveField(model_name="manifest", name="site_name"),
    ]



from django.db import models
from edc_search.model_mixins import SearchSlugManager
from edc_visit_tracking.managers import CrfModelManager


class AliquotManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, aliquot_identifier):
        return self.get(aliquot_identifier=aliquot_identifier)


class ManifestManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, manifest_identifier):
        return self.get(manifest_identifier=manifest_identifier)


class RequisitionManager(CrfModelManager, SearchSlugManager, models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, requisition_identifier):
        return self.get(requisition_identifier=requisition_identifier)



from __future__ import annotations

from typing import TYPE_CHECKING

from django import forms
from edc_utils import formatted_datetime, to_utc

if TYPE_CHECKING:
    from ..model_mixins import RequisitionModelMixin


class CrfRequisitionFormValidatorMixin:

    """An FormValidator mixin for CRFs (not requisitions).

    Used with a CRF that refers to a requisition or requisitions.

    Call in FormValidator.clean.

    For test 'xxx' expects the field trio of 'requisition' and
    'assay_datetime', 'xxx_panel'

        self.required_if_true(
            self.cleaned_data.get('cd4') is not None,
            field_required='cd4_requisition')
        self.validate_requisition(
            'cd4_requisition', 'cd4_assay_datetime', cd4_panel)

    See also, for example: ambition_form_validators.form_validators.blood_result
    """

    assay_datetime_field: str = "assay_datetime"
    requisition_field: str = "requisition"

    def validate_requisition(
        self,
        *panels,
        requisition_field: str | None = None,
        assay_datetime_field: str | None = None,
    ) -> RequisitionModelMixin:
        """Validates that the requisition model instance exists
        and assay datetime provided.
        """
        requisition_field = requisition_field or self.requisition_field
        requisition = self.cleaned_data.get(requisition_field)
        assay_datetime_field = assay_datetime_field or self.assay_datetime_field
        if requisition and requisition.panel_object not in panels:
            raise forms.ValidationError(
                {
                    requisition_field: (
                        "Incorrect requisition. Unknown panel. "
                        f"Expected one of {[p.name for p in panels]}. "
                        f"Got {requisition.panel_object.name}"
                    )
                }
            )

        self.required_if_true(requisition, field_required=assay_datetime_field)

        self.validate_assay_datetime(requisition, assay_datetime_field)
        return requisition

    def validate_assay_datetime(
        self, requisition: RequisitionModelMixin, assay_datetime_field: str | None = None
    ) -> None:
        assay_datetime_field = assay_datetime_field or self.assay_datetime_field
        assay_datetime = self.cleaned_data.get(assay_datetime_field)
        if assay_datetime:
            assay_datetime = to_utc(assay_datetime)
            requisition_datetime = to_utc(requisition.requisition_datetime)
            if assay_datetime < requisition_datetime:
                raise forms.ValidationError(
                    {
                        assay_datetime_field: (
                            f"Invalid. Cannot be before date of requisition "
                            f"{formatted_datetime(requisition_datetime)}."
                        )
                    }
                )



from django.db import models
from django.db.models.deletion import PROTECT
from edc_model import models as edc_models
from edc_model.validators import datetime_not_future
from edc_sites.models import CurrentSiteManager, SiteModelMixin
from edc_utils import get_utcnow

from .aliquot import Aliquot


class OrderManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, report_datetime, aliquot_identifier):
        return self.get(
            report_datetime=report_datetime,
            aliquot__aliquot_identifier=aliquot_identifier,
        )


class Order(SiteModelMixin, edc_models.BaseUuidModel):

    aliquot = models.ForeignKey(Aliquot, on_delete=PROTECT)

    order_identifier = models.CharField(max_length=25, editable=False, unique=True)

    order_datetime = models.DateTimeField(default=get_utcnow, validators=[datetime_not_future])

    panel_name = models.CharField(max_length=25)

    on_site = CurrentSiteManager()

    objects = OrderManager()

    history = edc_models.HistoricalRecords()

    def natural_key(self):
        return (self.report_datetime,) + self.aliquot.natural_key()

    natural_key.dependencies = ["edc_lab.aliquot", "sites.Site"]

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Order"



from django.db import models
from django.db.models.deletion import PROTECT
from edc_model import models as edc_models
from edc_sites.models import CurrentSiteManager

from ..model_mixins import ResultModelMixin
from .order import Order


class ResultManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, report_datetime, order_identifier):
        return self.get(
            report_datetime=report_datetime, order__order_identifier=order_identifier
        )


class Result(ResultModelMixin, edc_models.BaseUuidModel):

    order = models.ForeignKey(Order, on_delete=PROTECT)

    on_site = CurrentSiteManager()

    objects = ResultManager()

    history = edc_models.HistoricalRecords()

    def natural_key(self):
        return (self.report_datetime, self.order.order_identifier)

    natural_key.dependencies = ["edc_lab.order", "edc_lab.panel", "sites.Site"]

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Result"



from django.db import models
from edc_constants.constants import OPEN
from edc_utils import get_utcnow

from ...constants import VERIFIED


class VerifyModelMixin(models.Model):

    verified = models.IntegerField(default=0)

    verified_datetime = models.DateTimeField(null=True)

    def unverify(self):
        self.verified = 0
        self.verified_datetime = None
        self.save()
        self.box.save()

    class Meta:
        abstract = True


class VerifyBoxModelMixin(VerifyModelMixin, models.Model):
    def update_verified(self):
        if self.status in [OPEN, VERIFIED]:
            if self.is_verified:
                self.verified = 1
                self.status = VERIFIED
                self.verified_datetime = get_utcnow()
            else:
                self.verified = 0
                self.verified_datetime = None
                self.status = OPEN

    def unverify_box(self):
        if self.status in [OPEN, VERIFIED]:
            for box_item in self.boxitem_set.all():
                box_item.unverify()
                self.save()

    @property
    def is_verified(self):
        if self.boxitem_set.all().count() == 0:
            return False
        elif self.boxitem_set.filter(verified=False).exists():
            return False
        return True

    class Meta:
        abstract = True



from django.apps import apps as django_apps
from edc_constants.constants import YES

from ..identifiers import Prefix as IdentifierPrefix
from .aliquot_creator import AliquotCreator
from .primary_aliquot import PrimaryAliquot
from .specimen_processor import SpecimenProcessor


class SpecimenNotDrawnError(Exception):
    pass


class Specimen:

    """A class that represents a collected specimen and it's aliquots
    given the original requisition.

    The primary aliquot will be created if it does not already exist.
    """

    aliquot_creator_cls = AliquotCreator
    aliquot_model = "edc_lab.aliquot"
    identifier_prefix_cls = IdentifierPrefix
    primary_aliquot_cls = PrimaryAliquot
    specimen_processor_cls = SpecimenProcessor

    def __init__(self, requisition=None):

        self.requisition = requisition

        if not self.requisition.is_drawn == YES:
            raise SpecimenNotDrawnError(f"Specimen not drawn. Got '{requisition}'")

        self.aliquot_type = self.requisition.panel_object.aliquot_type

        if not self.requisition.identifier_prefix:
            self.requisition.identifier_prefix = self.primary_aliquot.identifier_prefix
            self.requisition.primary_aliquot_identifier = (
                self.primary_aliquot.aliquot_identifier
            )
            self.requisition.save()

    @property
    def primary_aliquot(self):
        """Returns a primary aliquot model instance after
        getting or creating one.
        """
        primary_aliquot_obj = self.primary_aliquot_cls(
            aliquot_creator_cls=self.aliquot_creator_cls,
            aliquot_type=self.aliquot_type,
            identifier_prefix=self.identifier_prefix,
            requisition_identifier=self.requisition.requisition_identifier,
            subject_identifier=self.requisition.subject_identifier,
        )
        return primary_aliquot_obj.object

    def process(self):
        """Returns a list of aliquots after getting or creating
        the aliquots according to the processing profile.
        """
        specimen_processor = self.specimen_processor_cls(
            aliquot_creator_cls=self.aliquot_creator_cls,
            identifier_prefix=self.identifier_prefix,
            model_obj=self.primary_aliquot,
            processing_profile=self.requisition.panel_object.processing_profile,
        )
        return specimen_processor.create()

    @property
    def aliquots(self):
        aliquot_model_cls = django_apps.get_model(self.aliquot_model)
        return aliquot_model_cls.objects.filter(identifier_prefix=self.identifier_prefix)

    @property
    def identifier_prefix(self):
        """Returns an identifier prefix string based on the
        requisition_identifier.
        """
        return self.identifier_prefix_cls(
            protocol_number=self.requisition.get_protocol_number(),
            requisition_identifier=self.requisition.requisition_identifier,
        )



from django.db import transaction
from django.db.utils import IntegrityError


class SpecimenProcessorError(Exception):
    pass


class SpecimenProcessor:
    """A class to process a specimen into aliquots according
    to its processing profile.
    """

    def __init__(
        self,
        model_obj=None,
        processing_profile=None,
        identifier_prefix=None,
        aliquot_creator_cls=None,
    ):
        self.aliquot_creator_cls = aliquot_creator_cls
        self.model_obj = model_obj
        self.processing_profile = processing_profile
        self.identifier_prefix = identifier_prefix

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"processing_profile={self.processing_profile},"
            f"identifier_prefix={self.identifier_prefix})"
        )

    def create(self):
        """Creates all aliquots in the processing profile "processes"."""
        created = []
        count = 1
        for process in self.processing_profile.processes.values():
            aliquot_count = process.aliquot_count
            for _ in range(1, aliquot_count + 1):
                count += 1
                aliquot_creator = self.aliquot_creator_cls(
                    parent_identifier=self.model_obj.aliquot_identifier,
                    requisition_identifier=self.model_obj.requisition_identifier,
                    subject_identifier=self.model_obj.subject_identifier,
                    identifier_prefix=self.identifier_prefix,
                )
                with transaction.atomic():
                    try:
                        aliquot = aliquot_creator.create(
                            count=count, aliquot_type=process.aliquot_type
                        )
                    except IntegrityError:
                        pass
                    else:
                        created.append(aliquot)
        return created



# Generated by Django 2.1.2 on 2018-11-08 01:53

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_lab", "0016_auto_20180706_1519")]

    operations = [
        migrations.AlterField(
            model_name="historicalaliquot",
            name="history_id",
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="historicalbox",
            name="history_id",
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="historicalboxitem",
            name="history_id",
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="history_id",
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="history_id",
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="history_id",
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="historicalresult",
            name="history_id",
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="historicalresultitem",
            name="history_id",
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, primary_key=True, serialize=False
            ),
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="history_id",
            field=models.UUIDField(
                default=uuid.uuid4, editable=False, primary_key=True, serialize=False
            ),
        ),
    ]



from django.test import TestCase

from edc_lab.lab import (
    AliquotType,
    AliquotTypeAlphaCodeError,
    AliquotTypeNumericCodeError,
)


class TestAliquotType(TestCase):
    def setUp(self):
        self.wb = AliquotType(name="whole_blood", numeric_code="02", alpha_code="WB")
        self.bc = AliquotType(name="buffy_coat", numeric_code="12", alpha_code="BC")

    def test_bad_aliquot_type1(self):
        self.assertRaises(
            AliquotTypeNumericCodeError,
            AliquotType,
            name="whole_blood",
            numeric_code="AA",
            alpha_code="WB",
        )

    def test_bad_aliquot_type2(self):
        self.assertRaises(
            AliquotTypeAlphaCodeError,
            AliquotType,
            name="whole_blood",
            numeric_code="02",
            alpha_code="99",
        )

    def test_aliquot_type_repr(self):
        aliquot_type = AliquotType(name="aliquot", numeric_code="00", alpha_code="AA")
        self.assertTrue(repr(aliquot_type))

    def test_aliquot_type_derivatives_single(self):
        """Asserts can add a derivative."""
        self.wb.add_derivatives(self.bc)
        self.assertEqual(self.wb.derivatives, [self.bc])

    def test_aliquot_type_derivatives_multi(self):
        """Asserts can add more than one derivative."""
        pl = AliquotType(name="plasma", numeric_code="32", alpha_code="PL")
        self.wb.add_derivatives(self.bc, pl)
        self.assertEqual(self.wb.derivatives, [self.bc, pl])



import sys

from django.apps import AppConfig as DjangoAppConfig
from django.conf import settings
from django.core.management.color import color_style
from django.db.models.signals import post_migrate
from edc_visit_tracking.constants import SCHEDULED

from .site_labs import site_labs

style = color_style()


def update_panels_on_post_migrate(sender, **kwargs):
    site_labs.migrated = True
    site_labs.autodiscover()
    if site_labs.loaded:
        site_labs.update_panel_model(panel_model_cls=sender.get_model("panel"), sender=sender)


class AppConfig(DjangoAppConfig):
    name = "edc_lab"
    verbose_name = "Edc Lab"
    has_exportable_data = True
    include_in_administration_section = True
    result_model = "edc_lab.result"

    def ready(self):
        from .models.signals import manifest_item_on_post_delete  # noqa

        post_migrate.connect(update_panels_on_post_migrate, sender=self)
        sys.stdout.write(f"Loading {self.verbose_name} ...\n")
        site_labs.autodiscover()
        sys.stdout.write(f" Done loading {self.verbose_name}.\n")


if settings.APP_NAME == "edc_lab":
    from edc_metadata.apps import AppConfig as EdcMetadataAppConfigBase

    class EdcMetadataAppConfig(EdcMetadataAppConfigBase):
        reason_field = {"edc_lab.subjectvisit": "reason"}
        create_on_reasons = [SCHEDULED]



from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from edc_protocol import Protocol

from ..site_labs import site_labs
from .base_label import BaseLabel


class AliquotLabelError(Exception):
    pass


class AliquotLabel(BaseLabel):
    model = "edc_lab.aliquot"
    template_name = "aliquot"
    registered_subject_model = "edc_registration.registeredsubject"

    def __init__(self, pk=None, children_count=None, request=None):
        super().__init__(pk=pk, request=request)
        self._registered_subject = None
        self._requisition = None
        self.children_count = children_count

    @property
    def requisition(self):
        """Returns a requisition model instance for this
        requisition_identifier."""
        if not self._requisition:
            for model in site_labs.requisition_models:
                model_cls = django_apps.get_model(model)
                try:
                    self._requisition = model_cls.objects.get(
                        requisition_identifier=self.model_obj.requisition_identifier
                    )
                except ObjectDoesNotExist:
                    pass
                else:
                    break
            if not self._requisition:
                raise AliquotLabelError(
                    "Requisition does not exist. "
                    f"requisition_identifier={self.model_obj.requisition_identifier}. "
                    f"Tried models {site_labs.requisition_models}"
                )
        return self._requisition

    @property
    def label_context(self):
        registered_subject = django_apps.get_model(self.registered_subject_model).objects.get(
            subject_identifier=self.requisition.subject_identifier
        )
        return {
            "aliquot_identifier": self.model_obj.human_readable_identifier,
            "aliquot_count": 1 if self.model_obj.is_primary else self.model_obj.count,
            "children_count": 1 if self.model_obj.is_primary else self.children_count,
            "primary": "<P>" if self.model_obj.is_primary else "",
            "barcode_value": self.model_obj.aliquot_identifier,
            "protocol": Protocol().protocol,
            "site": str(self.requisition.site.id),
            "site_name": str(self.requisition.site.name),
            "site_title": str(self.requisition.site.siteprofile.title),
            "clinician_initials": self.requisition.user_created[0:2].upper(),
            "drawn_datetime": self.requisition.drawn_datetime.strftime("%Y-%m-%d %H:%M"),
            "subject_identifier": registered_subject.subject_identifier,
            "gender": registered_subject.gender,
            "dob": registered_subject.dob.strftime("%Y-%m-%d"),
            "initials": registered_subject.initials,
            "alpha_code": self.model_obj.alpha_code,
            "panel": self.requisition.panel_object.abbreviation,
            "panel_name": self.requisition.panel.display_name,
        }



from django.db import models
from django.db.models.deletion import PROTECT

from ..site_labs import site_labs


class PanelModelError(Exception):
    pass


class LabProfileError(Exception):
    pass


class NothingPanel:
    verbose_name = None


class PanelModelMixin(models.Model):

    panel = models.ForeignKey("edc_lab.Panel", on_delete=PROTECT, null=True)

    @property
    def panel_object(self):
        """Returns a `panel` or `panel group` object"""
        try:
            panel_name = self.panel.name
        except AttributeError:
            panel_object = NothingPanel()
        else:
            try:
                panel_object = self.lab_profile_object.panels[panel_name]
            except KeyError as e:
                raise PanelModelError(
                    "Undefined `panel` name or `panel group` name. "
                    f"Got {panel_name}. See LabProfile and model Panel. Got {e}"
                )
        return panel_object

    @property
    def lab_profile_object(self):
        lab_profile_object = site_labs.get(self.panel.lab_profile_name)
        if not lab_profile_object:
            raise LabProfileError(
                f"Undefined lab profile name detected from panel {self.panel}. "
                f"Expected one of {site_labs.lab_profiles}. "
                f"Got '{self.panel.lab_profile_name}'. "
                "See stored values in panel model."
            )
        return lab_profile_object

    class Meta:
        abstract = True



from django.contrib import admin
from django_audit_fields.admin import ModelAdminAuditFieldsMixin, audit_fields
from django_revision.modeladmin_mixin import ModelAdminRevisionMixin
from edc_fieldsets import FieldsetsModelAdminMixin
from edc_model_admin.mixins import (
    ModelAdminFormAutoNumberMixin,
    ModelAdminFormInstructionsMixin,
    ModelAdminNextUrlRedirectMixin,
)


class BaseModelAdmin(
    ModelAdminFormInstructionsMixin,
    ModelAdminNextUrlRedirectMixin,
    ModelAdminFormAutoNumberMixin,
    ModelAdminRevisionMixin,
    ModelAdminAuditFieldsMixin,
    FieldsetsModelAdminMixin,
    admin.ModelAdmin,
):

    list_per_page = 10
    date_hierarchy = "modified"
    empty_value_display = "-"
    view_on_site = False
    show_cancel = True

    def get_readonly_fields(self, request, obj=None) -> tuple:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        return tuple(set(readonly_fields + audit_fields + ("site",)))



import re

from django.db import models
from django.db.models.deletion import PROTECT
from edc_model import models as edc_models
from edc_search.model_mixins import SearchSlugManager, SearchSlugModelMixin

from ..model_mixins import VerifyModelMixin
from ..patterns import aliquot_pattern
from .box import Box


class BoxItemManager(SearchSlugManager, models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, position, identifier, box_identifier):
        return self.get(
            position=position, identifier=identifier, box_identifier=box_identifier
        )


class BoxItem(SearchSlugModelMixin, VerifyModelMixin, edc_models.BaseUuidModel):

    box = models.ForeignKey(Box, on_delete=PROTECT)

    position = models.IntegerField()

    identifier = models.CharField(max_length=25)

    comment = models.CharField(max_length=25, null=True, blank=True)

    objects = BoxItemManager()

    history = edc_models.HistoricalRecords()

    def natural_key(self):
        return (self.position, self.identifier) + self.box.natural_key()

    natural_key.dependencies = ["edc_lab.box", "edc_lab.boxtype", "sites.Site"]

    @property
    def human_readable_identifier(self):
        """Returns a human-readable identifier"""
        if self.identifier:
            x = self.identifier
            if re.match(aliquot_pattern, self.identifier):
                return "{}-{}-{}-{}-{}".format(x[0:3], x[3:6], x[6:10], x[10:14], x[14:18])
        return self.identifier

    def get_slugs(self):
        slugs = [self.identifier, self.human_readable_identifier]
        return slugs

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Box Item"
        ordering = ("position",)
        unique_together = (("box", "position"), ("box", "identifier"))



import re

from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.test import TestCase
from edc_appointment.models import Appointment
from edc_constants.constants import NO, NOT_APPLICABLE, YES
from edc_facility import import_holidays
from edc_sites import add_or_update_django_sites
from edc_sites.single_site import SingleSite
from edc_utils.date import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED

from edc_lab.lab import (
    AliquotType,
    LabProfile,
    Process,
    ProcessingProfile,
    ProcessingProfileAlreadyAdded,
)
from edc_lab.site_labs import SiteLabs, site_labs

from ..models import SubjectConsent, SubjectRequisition, SubjectVisit
from ..site_labs_test_helper import SiteLabsTestHelper
from ..visit_schedules import visit_schedule


class TestSiteLab(TestCase):
    @classmethod
    def setUpClass(cls):
        add_or_update_django_sites(
            sites=[
                SingleSite(
                    settings.SITE_ID,
                    "test_site",
                    country_code="ug",
                    country="uganda",
                    domain="bugamba.ug.clinicedc.org",
                )
            ]
        )
        return super().setUpClass()

    def test_site_labs(self):
        site_lab = SiteLabs()
        self.assertFalse(site_lab.loaded)

    def test_site_labs_register(self):
        lab_profile = LabProfile(
            name="lab_profile", requisition_model="edc_lab.subjectrequisition"
        )
        site_lab = SiteLabs()
        site_lab.register(lab_profile)
        self.assertTrue(site_lab.loaded)

    def test_site_labs_register_none(self):
        site_lab = SiteLabs()
        site_lab.register(None)
        self.assertFalse(site_lab.loaded)


class TestSiteLab2(TestCase):

    lab_helper = SiteLabsTestHelper()

    @classmethod
    def setUpClass(cls):
        add_or_update_django_sites(
            sites=[
                SingleSite(
                    settings.SITE_ID,
                    "test_site",
                    country_code="ug",
                    country="uganda",
                    domain="bugamba.ug.clinicedc.org",
                )
            ]
        )
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)
        self.lab_helper.setup_site_labs()
        self.panel = self.lab_helper.panel
        self.lab_profile = self.lab_helper.lab_profile

        self.subject_identifier = "1111111111"
        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow(),
            identity="1111111",
            confirm_identity="1111111",
            visit_schedule_name="visit_schedule",
            schedule_name="schedule",
            dob=get_utcnow() - relativedelta(years=25),
        )
        appointment = Appointment.objects.get(visit_code="1000")
        self.subject_visit = SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )

    def test_site_lab_panels(self):
        self.assertIn(self.panel.name, site_labs.get(self.lab_profile.name).panels)

    def test_panel_repr(self):
        self.assertTrue(repr(self.panel))

    def test_assert_cannot_add_duplicate_process(self):
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        b = AliquotType(name="aliquot_b", numeric_code="66", alpha_code="BB")
        a.add_derivatives(b)
        process = Process(aliquot_type=b, aliquot_count=3)
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        processing_profile.add_processes(process)
        self.assertRaises(
            ProcessingProfileAlreadyAdded, processing_profile.add_processes, process
        )

    def test_requisition_specimen(self):
        """Asserts can create a requisition."""
        SubjectRequisition.objects.create(
            subject_visit=self.subject_visit, panel=self.panel.panel_model_obj
        )

    def test_requisition_identifier2(self):
        """Asserts requisition identifier is set on requisition."""
        requisition = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel.panel_model_obj,
            is_drawn=YES,
        )
        pattern = re.compile("[0-9]{2}[A-Z0-9]{5}")
        self.assertTrue(pattern.match(requisition.requisition_identifier))

    def test_requisition_identifier3(self):
        """Asserts requisition identifier is NOT set on requisition
        if specimen not drawn.
        """
        requisition = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel.panel_model_obj,
            is_drawn=NO,
            reason_not_drawn=NOT_APPLICABLE,
        )
        # is never None, even if not drawn.
        self.assertIsNotNone(requisition.requisition_identifier)
        # if not drawn, format is not UUID
        pattern = re.compile("^[0-9]{2}[A-Z0-9]{5}$")
        self.assertFalse(pattern.match(requisition.requisition_identifier))

    def test_requisition_identifier5(self):
        """Asserts requisition identifier is set if specimen
        changed to drawn.
        """
        requisition = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel.panel_model_obj,
            is_drawn=NO,
        )
        requisition.is_drawn = YES
        requisition.save()
        pattern = re.compile("[0-9]{2}[A-Z0-9]{5}")
        self.assertTrue(pattern.match(requisition.requisition_identifier))

    def test_requisition_identifier6(self):
        """Asserts requisition identifier is unchanged on save/resave."""
        requisition = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel.panel_model_obj,
            is_drawn=YES,
        )
        requisition_identifier = requisition.requisition_identifier
        requisition.is_drawn = YES
        requisition.save()
        self.assertEqual(requisition_identifier, requisition.requisition_identifier)



from dateutil.relativedelta import relativedelta
from edc_visit_schedule import Crf, FormsCollection, Schedule, Visit, VisitSchedule
from edc_visit_schedule.tests.dummy_panel import DummyPanel
from edc_visit_schedule.visit.requisition import Requisition


class MockPanel(DummyPanel):
    """`requisition_model` is normally set when the lab profile
    is set up.
    """

    def __init__(self, name):
        super().__init__(requisition_model="edc_lab.subjectrequisition", name=name)


crfs = FormsCollection(Crf(show_order=1, model="edc_lab.crfone", required=True))

requisitions = FormsCollection(
    Requisition(show_order=10, panel=MockPanel("panel"), required=True, additional=False)
)


visit = Visit(
    code="1000",
    timepoint=0,
    rbase=relativedelta(days=0),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    requisitions=requisitions,
    crfs=crfs,
    requisitions_unscheduled=None,
    crfs_unscheduled=None,
    allow_unscheduled=False,
    facility_name="5-day-clinic",
)


schedule = Schedule(
    name="schedule",
    onschedule_model="edc_lab.onschedule",
    offschedule_model="edc_lab.offschedule",
    appointment_model="edc_appointment.appointment",
    consent_model="edc_lab.subjectconsent",
)

visit_schedule = VisitSchedule(
    name="visit_schedule",
    offstudy_model="edc_offstudy.subjectoffstudy",
    death_report_model="edc_appointment.deathreport",
    locator_model="edc_locator.subjectlocator",
)

schedule.add_visit(visit)

visit_schedule.add_schedule(schedule)



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_lab_admin
from ..models import Panel
from .base_model_admin import BaseModelAdmin


@admin.register(Panel, site=edc_lab_admin)
class PanelAdmin(BaseModelAdmin, admin.ModelAdmin):

    fieldsets = (
        (None, {"fields": ("name", "display_name", "lab_profile_name")}),
        audit_fieldset_tuple,
    )

    def get_readonly_fields(self, request, obj=None) -> tuple:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        return readonly_fields + ("name", "display_name", "lab_profile_name")



from django.contrib import admin
from edc_model_admin.mixins import TabularInlineMixin

from ..admin_site import edc_lab_admin
from ..models import Result, ResultItem
from .base_model_admin import BaseModelAdmin


@admin.register(ResultItem, site=edc_lab_admin)
class ResultItemAdmin(BaseModelAdmin, admin.ModelAdmin):
    pass


class ResultItemInlineAdmin(TabularInlineMixin, admin.TabularInline):
    model = ResultItem


@admin.register(Result, site=edc_lab_admin)
class ResultAdmin(BaseModelAdmin, admin.ModelAdmin):
    inlines = [ResultItemInlineAdmin]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-05 17:39
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_lab", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="consignee",
            name="state",
            field=models.CharField(
                blank=True, max_length=50, null=True, verbose_name="State or Province"
            ),
        ),
        migrations.AddField(
            model_name="historicalconsignee",
            name="state",
            field=models.CharField(
                blank=True, max_length=50, null=True, verbose_name="State or Province"
            ),
        ),
        migrations.AddField(
            model_name="historicalshipper",
            name="state",
            field=models.CharField(
                blank=True, max_length=50, null=True, verbose_name="State or Province"
            ),
        ),
        migrations.AddField(
            model_name="shipper",
            name="state",
            field=models.CharField(
                blank=True, max_length=50, null=True, verbose_name="State or Province"
            ),
        ),
        migrations.AlterField(
            model_name="consignee",
            name="address",
            field=models.CharField(default="address", max_length=50, verbose_name="Address"),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="consignee",
            name="city",
            field=models.CharField(default="city", max_length=50),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="consignee",
            name="country",
            field=models.CharField(default="country", max_length=50),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="consignee",
            name="postal_code",
            field=models.CharField(default="0000", max_length=50),
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="address",
            field=models.CharField(default="address", max_length=50, verbose_name="Address"),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="city",
            field=models.CharField(default="city", max_length=50),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="country",
            field=models.CharField(default="country", max_length=50),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="postal_code",
            field=models.CharField(default="0000", max_length=50),
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="address",
            field=models.CharField(default="address", max_length=50, verbose_name="Address"),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="city",
            field=models.CharField(default="city", max_length=50),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="country",
            field=models.CharField(default="country", max_length=50),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="postal_code",
            field=models.CharField(default="0000", max_length=50),
        ),
        migrations.AlterField(
            model_name="shipper",
            name="address",
            field=models.CharField(default="address", max_length=50, verbose_name="Address"),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="shipper",
            name="city",
            field=models.CharField(default="city", max_length=50),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="shipper",
            name="country",
            field=models.CharField(default="country", max_length=50),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="shipper",
            name="postal_code",
            field=models.CharField(default="0000", max_length=50),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-21 09:19
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_lab", "0006_auto_20170305_2121")]

    operations = [
        migrations.AlterField(
            model_name="consignee",
            name="name",
            field=models.CharField(help_text="Company name", max_length=50, unique=True),
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="name",
            field=models.CharField(db_index=True, help_text="Company name", max_length=50),
        ),
    ]



import re

from django.conf import settings
from django.test import TestCase
from edc_sites import add_or_update_django_sites
from edc_sites.single_site import SingleSite

from edc_lab.identifiers import RequisitionIdentifier
from edc_lab.lab import (
    AliquotType,
    LabProfile,
    Process,
    ProcessingProfile,
    RequisitionPanel,
)
from edc_lab.site_labs import site_labs


class TestRequisition(TestCase):
    @classmethod
    def setUpClass(cls):
        add_or_update_django_sites(
            sites=[
                SingleSite(
                    settings.SITE_ID,
                    "test_site",
                    country_code="ug",
                    country="uganda",
                    domain="bugamba.ug.clinicedc.org",
                )
            ]
        )
        return super().setUpClass()

    def test_requisition_identifier(self):
        """Asserts requisition identifier class creates identifier
        with correct format.
        """
        identifier = RequisitionIdentifier()
        pattern = re.compile("[0-9]{2}[A-Z0-9]{5}")
        self.assertTrue(pattern.match(str(identifier)))


class TestRequisitionModel(TestCase):
    @classmethod
    def setUpClass(cls):
        add_or_update_django_sites(
            sites=[
                SingleSite(
                    settings.SITE_ID,
                    "test_site",
                    country_code="ug",
                    country="uganda",
                    domain="bugamba.ug.clinicedc.org",
                )
            ]
        )
        return super().setUpClass()

    def setUp(self):
        self.requisition_model = "edc_lab.subjectrequisition"
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        b = AliquotType(name="aliquot_b", numeric_code="66", alpha_code="BB")
        a.add_derivatives(b)
        process = Process(aliquot_type=b, aliquot_count=3)
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        processing_profile.add_processes(process)
        panel = RequisitionPanel(name="Viral Load", processing_profile=processing_profile)
        self.lab_profile = LabProfile(name="profile", requisition_model=self.requisition_model)
        self.lab_profile.add_panel(panel)
        site_labs._registry = {}
        site_labs.loaded = False
        site_labs.register(lab_profile=self.lab_profile)

    def test_(self):
        obj = site_labs.get(lab_profile_name="profile")
        self.assertEqual(obj, self.lab_profile)

    def test_lab_profile_model(self):
        obj = site_labs.get(lab_profile_name="profile")
        self.assertEqual(self.requisition_model, obj.requisition_model)

    def test_panel_model(self):
        for panel in site_labs.get(lab_profile_name="profile").panels.values():
            self.assertEqual(panel.requisition_model, self.requisition_model)



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_lab_admin
from ..forms import BoxItemForm
from ..models import BoxItem
from .base_model_admin import BaseModelAdmin


@admin.register(BoxItem, site=edc_lab_admin)
class BoxItemAdmin(BaseModelAdmin, admin.ModelAdmin):

    form = BoxItemForm

    fieldsets = (
        (None, {"fields": ("box", "position", "identifier", "comment")}),
        audit_fieldset_tuple,
    )

    def get_list_filter(self, request) -> tuple:
        list_filter = super().get_list_filter(request)
        custom_fields = ("identifier", "position")
        return tuple(set(custom_fields + list_filter))



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_lab_admin
from ..forms import BoxTypeForm
from ..models import BoxType
from .base_model_admin import BaseModelAdmin


@admin.register(BoxType, site=edc_lab_admin)
class BoxTypeAdmin(BaseModelAdmin, admin.ModelAdmin):

    form = BoxTypeForm

    fieldsets = (
        (None, {"fields": ("name", "across", "down", "total", "fill_order")}),
        audit_fieldset_tuple,
    )

    def get_list_filter(self, request) -> tuple:
        list_filter = super().get_list_filter(request)
        custom_fields = ("name", "across", "down", "total")
        return tuple(set(custom_fields + list_filter))



from django.conf import settings
from django.db.utils import IntegrityError
from django.test import TestCase
from edc_sites import add_or_update_django_sites
from edc_sites.single_site import SingleSite
from edc_sites.tests import SiteTestCaseMixin

from edc_lab.lab import AliquotCreator, AliquotCreatorError
from edc_lab.models import Aliquot


class TestAliquot(SiteTestCaseMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        add_or_update_django_sites(
            sites=[
                SingleSite(
                    settings.SITE_ID,
                    "test_site",
                    country_code="ug",
                    country="uganda",
                    domain="bugamba.ug.clinicedc.org",
                )
            ]
        )
        return super().setUpClass()

    def test_aliquot_model_constraint(self):
        Aliquot.objects.create(count=0)
        self.assertRaises(IntegrityError, Aliquot.objects.create, count=0)

    def test_create_aliquot(self):
        self.assertRaises(AliquotCreatorError, AliquotCreator)



from tempfile import mkdtemp

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from edc_pdf_reports import Report
from reportlab.graphics.barcode import code39
from reportlab.lib import colors
from reportlab.lib.units import cm, mm
from reportlab.platypus import Paragraph, Spacer, Table, TableStyle

from ..models import Aliquot, Box, BoxItem
from ..site_labs import site_labs


class ManifestReportError(Exception):
    def __init__(self, message, code=None):
        super().__init__(message)
        self.code = code


class ManifestReport(Report):
    def __init__(self, manifest=None, user=None, **kwargs):
        super().__init__(**kwargs)
        self.manifest = manifest  # a Manifest model instance
        self.user = user  # a User model instance
        self.box_model = Box
        self.box_item_model = BoxItem
        self.aliquot_model = Aliquot
        self.image_folder = mkdtemp()
        if self.manifest.shipped:
            self.report_filename = f"{self.manifest.manifest_identifier}.pdf"
        else:
            self.report_filename = "manifest_preview.pdf"

    def get_report_story(self, **kwargs):
        story = []

        data = [
            [
                Paragraph(
                    " ".join(self.manifest.site.name.split("_")).upper(),
                    self.styles["line_data_large"],
                )
            ],
            [Paragraph("SITE NAME", self.styles["line_label"])],
            [
                Paragraph(
                    self.formatted_address(**self.manifest.shipper.__dict__),
                    self.styles["line_data_large"],
                )
            ],
            [Paragraph("SITE DETAILS", self.styles["line_label"])],
        ]

        t = Table(data, colWidths=(9 * cm))
        t.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (0, 1), 0.25, colors.black),
                    ("INNERGRID", (0, 2), (0, 3), 0.25, colors.black),
                ]
            )
        )
        t.hAlign = "RIGHT"

        story.append(t)

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(
            Paragraph(
                "SPECIMEN MANIFEST{}".format(" (reprint)" if self.manifest.printed else ""),
                self.styles["line_label_center"],
            )
        )

        if self.manifest.shipped:
            barcode = code39.Standard39(
                self.manifest.manifest_identifier, barHeight=10 * mm, stop=1
            )
        else:
            barcode = "PREVIEW"

        data = [
            [
                Paragraph("MANIFEST NO.", self.styles["line_label"]),
                Paragraph(
                    (
                        self.manifest.human_readable_identifier
                        if self.manifest.shipped
                        else "PREVIEW"
                    ),
                    self.styles["line_data_largest"],
                ),
                barcode,
            ]
        ]
        t = Table(data, colWidths=(3 * cm, None, 5.5 * cm))
        t.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (1, 0), (-1, -1), 0.25, colors.black),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )

        story.append(t)

        data = [
            [
                Paragraph("MANIFEST DATE", self.styles["line_label"]),
                Paragraph("EXPORT REFERENCES", self.styles["line_label"]),
            ],
            [
                Paragraph(
                    self.manifest.manifest_datetime.strftime("%Y-%m-%d"),
                    self.styles["line_data_largest"],
                ),
                Paragraph(
                    self.manifest.export_references or "",
                    self.styles["line_data_largest"],
                ),
            ],
        ]
        t = Table(data)
        t.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (1, 0), 0.25, colors.black),
                    ("INNERGRID", (0, 1), (1, 1), 0.25, colors.black),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t)

        data = [
            [
                Paragraph(
                    "SHIPPER/EXPORTER (complete name and address)",
                    self.styles["line_label"],
                ),
                Paragraph("CONSIGNEE (complete name and address)", self.styles["line_label"]),
            ],
            [
                Paragraph(
                    self.formatted_address(**self.shipper_data),
                    self.styles["line_data_large"],
                ),
                Paragraph(
                    self.formatted_address(**self.consignee_data),
                    self.styles["line_data_large"],
                ),
            ],
        ]

        t = Table(data)
        t.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (1, 0), 0.25, colors.black),
                    ("INNERGRID", (0, 1), (1, 1), 0.25, colors.black),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t)

        data1 = [
            [
                Paragraph("COUNTRY OF EXPORT", self.styles["line_label"]),
                Paragraph(
                    "DESCRIPTION OF GOODS (description and special instructions)<br />",
                    self.styles["line_label"],
                ),
            ],
            [
                Paragraph(self.manifest.shipper.country, self.styles["line_data_largest"]),
                Paragraph(self.description, self.styles["line_data_large"]),
            ],
            [Paragraph("COUNTRY OF ORIGIN", self.styles["line_label"]), ""],
            [
                Paragraph(self.manifest.shipper.country, self.styles["line_data_largest"]),
                "",
            ],
            [
                Paragraph("COUNTRY OF ULTIMATE DESTINATION", self.styles["line_label"]),
                "",
            ],
            [
                Paragraph(self.manifest.consignee.country, self.styles["line_data_largest"]),
                "",
            ],
        ]
        t1 = Table(data1)
        t1.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 1), (0, 2), 0.25, colors.black),
                    ("INNERGRID", (0, 3), (0, 4), 0.25, colors.black),
                    ("INNERGRID", (0, 0), (1, 0), 0.25, colors.black),
                    ("INNERGRID", (0, 1), (1, 1), 0.25, colors.black),
                    ("INNERGRID", (0, 2), (1, 2), 0.25, colors.black),
                    ("INNERGRID", (0, 3), (1, 3), 0.25, colors.black),
                    ("INNERGRID", (0, 4), (1, 4), 0.25, colors.black),
                    ("INNERGRID", (0, 5), (1, 5), 0.25, colors.black),
                    ("SPAN", (1, 1), (1, 5)),
                    ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                ]
            )
        )
        story.append(t1)

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        story.append(
            Table(
                [
                    [
                        Paragraph(
                            "I DECLARE THE INFORMATION CONTAINED IN THIS "
                            "MANIFEST TO BE TRUE AND CORRECT",
                            self.styles["line_label"],
                        )
                    ]
                ]
            )
        )

        story.append(Spacer(0.1 * cm, 0.5 * cm))

        data1 = [
            [
                Paragraph(self.contact_name, self.styles["line_data_large"]),
                "",
                Paragraph(
                    self.manifest.export_datetime.strftime("%Y-%m-%d %H:%M")
                    if self.manifest.shipped
                    else "PREVIEW",
                    self.styles["line_data_large"],
                ),
            ],
            [
                Paragraph(
                    "SIGNATURE OF SHIPPER/EXPORTER (Type name and title and sign.)",
                    self.styles["line_label"],
                ),
                "",
                Paragraph("DATE", self.styles["line_label"]),
            ],
        ]

        t1 = Table(data1, colWidths=(None, 2 * cm, None))
        t1.setStyle(
            TableStyle(
                [
                    ("INNERGRID", (0, 0), (0, 1), 0.25, colors.black),
                    ("INNERGRID", (2, 0), (2, 1), 0.25, colors.black),
                ]
            )
        )

        story.append(t1)
        story.append(Spacer(0.1 * cm, 0.5 * cm))

        # boxes
        story.append(
            Table([[Paragraph("MANIFEST CONTENTS", self.styles["line_label_center"])]])
        )

        story = self.append_manifest_items_story(story)

        if self.manifest.shipped and not self.manifest.printed:
            self.manifest.printed = True
            self.manifest.save()

        return story

    @property
    def contact_name(self):
        return f"{self.user.first_name} {self.user.last_name}"

    @property
    def shipper_data(self):
        data = self.manifest.shipper.__dict__
        if self.contact_name.strip():
            data.update(contact_name=self.contact_name)
        return data

    @property
    def consignee_data(self):
        data = self.manifest.consignee.__dict__
        return data

    def formatted_address(self, **kwargs):
        data = {
            "contact_name": None,
            "name": None,
            "address": None,
            "city": None,
            "state": None,
            "postal_code": "0000",
            "country": None,
        }
        data.update(**kwargs)
        data_list = [
            v
            for v in [
                data.get("contact_name"),
                data.get("name"),
                data.get("address"),
                data.get("city")
                if not data.get("state")
                else "{} {}".format(data.get("city"), data.get("state")),
                data.get("postal_code"),
                data.get("country"),
            ]
            if v
        ]
        return "<br />".join(data_list)

    @property
    def description(self):
        boxes = self.box_model.objects.filter(
            box_identifier__in=[obj.identifier for obj in self.manifest.manifestitem_set.all()]
        )
        box_items = self.box_item_model.objects.filter(box__in=boxes)
        aliquots = self.aliquot_model.objects.filter(
            aliquot_identifier__in=[obj.identifier for obj in box_items]
        )
        specimen_types = list(set([obj.aliquot_type for obj in aliquots]))
        description = {
            "box_count": boxes.count(),
            "specimen_count": aliquots.count(),
            "specimen_types": ", ".join(specimen_types),
        }
        box_word = "box" if boxes.count() == 1 else "boxes"
        specimen_word = "specimen" if aliquots.count() == 1 else "specimens"
        type_word = "type" if len(specimen_types) == 1 else "types"
        return (
            "{box_count} {box_word} containing {specimen_count} "
            "{specimen_word} of {type_word} {specimen_types}.".format(
                box_word=box_word,
                specimen_word=specimen_word,
                type_word=type_word,
                **description,
            )
        )

    def append_manifest_items_story(self, story):
        box_header = [
            Paragraph("BOX:", self.styles["line_label"]),
            Paragraph("CATEGORY:", self.styles["line_label"]),
            Paragraph("TYPES:", self.styles["line_label"]),
            Paragraph("ITEMS:", self.styles["line_label"]),
            Paragraph("BOX DATE:", self.styles["line_label"]),
            Paragraph("BOX BARCODE:", self.styles["line_label"]),
        ]
        for index, manifest_item in enumerate(
            self.manifest.manifestitem_set.all().order_by("-created")
        ):
            if index > 0:
                story.append(Spacer(0.1 * cm, 0.5 * cm))
            data1 = []
            data1.append(box_header)
            try:
                box = self.box_model.objects.get(box_identifier=manifest_item.identifier)
            except ObjectDoesNotExist as e:
                raise ManifestReportError(
                    f"{e} Got Manifest item '{manifest_item.identifier}'.",
                    code="unboxed_item",
                ) from e
            barcode = code39.Standard39(box.box_identifier, barHeight=5 * mm, stop=1)
            data1.append(
                [
                    Paragraph(box.box_identifier, self.styles["line_data_large"]),
                    Paragraph(
                        box.get_category_display().upper(),
                        self.styles["line_data_large"],
                    ),
                    Paragraph(box.specimen_types, self.styles["line_data_large"]),
                    Paragraph(
                        f"{str(box.count)}/{str(box.box_type.total)}",
                        self.styles["line_data_large"],
                    ),
                    Paragraph(
                        box.box_datetime.strftime("%Y-%m-%d"),
                        self.styles["line_data_large"],
                    ),
                    barcode,
                ]
            )
            t1 = Table(data1, colWidths=(None, None, None, None, None, None))
            t1.setStyle(
                TableStyle(
                    [
                        ("INNERGRID", (0, 0), (-1, 0), 0.25, colors.black),
                        ("INNERGRID", (0, 1), (-1, -1), 0.25, colors.black),
                        ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ]
                )
            )
            story.append(t1)
            story.append(Spacer(0.1 * cm, 0.5 * cm))

            table_data = []
            table_data = [
                [
                    Paragraph("BARCODE", self.styles["line_label_center"]),
                    Paragraph("POS", self.styles["line_label_center"]),
                    Paragraph("ALIQUOT IDENTIFIER", self.styles["line_label_center"]),
                    Paragraph("SUBJECT", self.styles["line_label_center"]),
                    Paragraph("TYPE", self.styles["line_label_center"]),
                    Paragraph("DATE", self.styles["line_label_center"]),
                ]
            ]
            for box_item in box.boxitem_set.all().order_by("position"):
                aliquot = self.get_aliquot(box_item.identifier)
                panel_object = self.get_panel_object(aliquot)
                barcode = code39.Standard39(
                    aliquot.aliquot_identifier, barHeight=5 * mm, stop=1
                )
                table_data.append(
                    [
                        barcode,
                        Paragraph(str(box_item.position), self.styles["row_data"]),
                        Paragraph(aliquot.human_readable_identifier, self.styles["row_data"]),
                        Paragraph(aliquot.subject_identifier, self.styles["row_data"]),
                        Paragraph(
                            "{} ({}) {}".format(
                                aliquot.aliquot_type,
                                aliquot.numeric_code,
                                panel_object.abbreviation,
                            ),
                            self.styles["row_data"],
                        ),
                        Paragraph(
                            aliquot.aliquot_datetime.strftime("%Y-%m-%d"),
                            self.styles["row_data"],
                        ),
                    ]
                )
            t1 = Table(table_data)
            t1.setStyle(
                TableStyle(
                    [
                        ("INNERGRID", (0, 0), (-1, -1), 0.25, colors.black),
                        ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),
                        ("BOX", (0, 0), (-1, -1), 0.25, colors.black),
                    ]
                )
            )
            story.append(t1)
        return story

    def get_aliquot(self, box_item_identifier=None):
        """Returns the aliquot instance for this box item."""
        try:
            aliquot = self.aliquot_model.objects.get(aliquot_identifier=box_item_identifier)
        except ObjectDoesNotExist as e:
            raise ManifestReportError(
                f"{e} Got Box item '{box_item_identifier}'",
                code="invalid_aliquot_identifier",
            )
        return aliquot

    def get_panel_object(self, aliquot=None):
        """Returns the panel object associated with this
        aliquot.
        """
        requisition = None
        for requisition_model in site_labs.requisition_models:
            model_cls = django_apps.get_model(requisition_model)
            try:
                requisition = model_cls.objects.get(
                    requisition_identifier=aliquot.requisition_identifier
                )
            except ObjectDoesNotExist:
                pass
        if not requisition:
            raise ManifestReportError(
                f"Invalid requisition identifier. Got {aliquot.requisition_identifier}",
                code="invalid_requisition_identifier",
            )
        return requisition.panel_object



from edc_lab.lab import (
    AliquotType,
    LabProfile,
    Process,
    ProcessingProfile,
    RequisitionPanel,
)
from edc_lab.site_labs import site_labs


class SiteLabsTestHelper:
    def __init__(self):
        self.profile_aliquot_count = None
        self.panel = None
        self.lab_profile = None

    requisition_model = "edc_lab.subjectrequisition"

    def setup_site_labs(self):
        """Sets up the site_lab global."""
        site_labs._registry = {}
        site_labs.loaded = False

        self.profile_aliquot_count = 3

        # create aliquots and their relationship
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        b = AliquotType(name="aliquot_b", numeric_code="66", alpha_code="BB")
        a.add_derivatives(b)

        # set up processes
        process = Process(aliquot_type=b, aliquot_count=self.profile_aliquot_count)
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        processing_profile.add_processes(process)

        # create a panel
        self.panel = RequisitionPanel(name="panel", processing_profile=processing_profile)

        # lab profile
        self.lab_profile = LabProfile(
            name="lab_profile", requisition_model=self.requisition_model
        )
        self.lab_profile.add_panel(self.panel)

        # register with site
        site_labs.register(lab_profile=self.lab_profile)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from django import forms
from django.apps import apps as django_apps
from edc_constants.constants import NO, YES
from edc_utils import to_utc

if TYPE_CHECKING:
    from ..models import Aliquot


class RequisitionFormValidatorError(Exception):
    pass


class RequisitionFormValidatorMixin:

    aliquot_model: str = "edc_lab.aliquot"

    def clean(self) -> None:
        if self.instance:
            if self.cleaned_data.get("packed") != self.instance.packed:
                raise forms.ValidationError({"packed": "Value may not be changed here."})
            elif self.cleaned_data.get("processed") != self.instance.processed:
                if self.aliqout_model_cls.objects.filter(
                    requisition_identifier=self.instance.requisition_identifier
                ).exists():
                    raise forms.ValidationError(
                        {"processed": "Value may not be changed. Aliquots exist."}
                    )
            elif not self.cleaned_data.get("received") and self.instance.received:
                if self.instance.processed:
                    raise forms.ValidationError(
                        {"received": "Specimen has already been processed."}
                    )
            elif self.cleaned_data.get("received") and not self.instance.received:
                raise forms.ValidationError(
                    {"received": "Receive specimens in the lab section of the EDC."}
                )
            elif self.instance.received:
                raise forms.ValidationError(
                    "Requisition may not be changed. The specimen has "
                    "already been received."
                )

        self.applicable_if(NO, field="is_drawn", field_applicable="reason_not_drawn")
        self.validate_other_specify(field="reason_not_drawn")
        self.required_if(YES, field="is_drawn", field_required="drawn_datetime")
        self.validate_drawn_datetime()
        self.applicable_if(YES, field="is_drawn", field_applicable="item_type")
        self.required_if(YES, field="is_drawn", field_required="item_count")
        self.required_if(YES, field="is_drawn", field_required="estimated_volume")

    @property
    def aliqout_model_cls(self) -> Aliquot:
        return django_apps.get_model(self.aliquot_model)

    @property
    def requisition_datetime(self) -> datetime | None:
        """Returns the requisition_datetime.

        `report_datetime` exists on the model but is always
        the same as `requisition_datetime`. See model.save().

        Note value of attr `report_datetime_field_attr`.
        """
        if self.report_datetime_field_attr != "requisition_datetime":
            raise RequisitionFormValidatorError(
                "Invalid value `report_datetime_field_attr`. Expected `requisition_datetime`. "
                f"Got `{self.report_datetime_field_attr}`."
            )
        return self.report_datetime

    def validate_drawn_datetime(self) -> None:
        if (
            self.requisition_datetime
            and self.cleaned_data.get("drawn_datetime")
            and to_utc(self.cleaned_data.get("drawn_datetime")).date()
            > to_utc(self.requisition_datetime).date()
        ):
            raise forms.ValidationError(
                {"drawn_datetime": "Invalid. Cannot be after requisition date."}
            )



from django.contrib import admin
from edc_model_admin.mixins import AddressModelAdminMixin

from ..admin_site import edc_lab_admin
from ..models import Consignee
from .base_model_admin import BaseModelAdmin


@admin.register(Consignee, site=edc_lab_admin)
class ConsigneeAdmin(BaseModelAdmin, AddressModelAdminMixin, admin.ModelAdmin):
    pass



# Generated by Django 2.0.1 on 2018-01-17 12:38

import django.db.models.deletion
import edc_sites.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("sites", "0002_alter_domain_unique"),
        ("edc_lab", "0012_auto_20180114_1438"),
    ]

    operations = [
        migrations.AlterModelOptions(name="order", options={}),
        migrations.AlterModelManagers(
            name="aliquot",
            managers=[("on_site", edc_sites.models.CurrentSiteManager())],
        ),
        migrations.AlterModelManagers(
            name="box", managers=[("on_site", edc_sites.models.CurrentSiteManager())]
        ),
        migrations.AlterModelManagers(
            name="manifest",
            managers=[("on_site", edc_sites.models.CurrentSiteManager())],
        ),
        migrations.AlterModelManagers(
            name="order", managers=[("on_site", edc_sites.models.CurrentSiteManager())]
        ),
        migrations.AlterModelManagers(
            name="result", managers=[("on_site", edc_sites.models.CurrentSiteManager())]
        ),
        migrations.AlterModelManagers(
            name="resultitem",
            managers=[("on_site", edc_sites.models.CurrentSiteManager())],
        ),
        migrations.AddField(
            model_name="aliquot",
            name="site",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="box",
            name="site",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="historicalaliquot",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="historicalbox",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="historicalmanifest",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="historicalorder",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="historicalresult",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="historicalresultitem",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="manifest",
            name="site",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="order",
            name="site",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="result",
            name="site",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="sites.Site",
            ),
        ),
        migrations.AddField(
            model_name="resultitem",
            name="site",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="sites.Site",
            ),
        ),
    ]



from .aliquot_identifier import (
    AliquotIdentifier,
    AliquotIdentifierCountError,
    AliquotIdentifierLengthError,
)
from .box_identifier import BoxIdentifier
from .manifest_identifier import ManifestIdentifier
from .prefix import Prefix, PrefixKeyError, PrefixLengthError
from .requisition_identifier import RequisitionIdentifier



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-05 19:09
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_lab", "0004_auto_20170305_2015")]

    operations = [
        migrations.AddField(
            model_name="historicalmanifest",
            name="printed",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="historicalmanifest",
            name="printed_datetime",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name="manifest",
            name="printed",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="manifest",
            name="printed_datetime",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="box",
            name="status",
            field=models.CharField(
                choices=[
                    ("open", "Open"),
                    ("verified", "Verified"),
                    ("packed", "Packed"),
                    ("shipped", "Shipped"),
                ],
                default="open",
                max_length=15,
            ),
        ),
        migrations.AlterField(
            model_name="consignee",
            name="contact_name",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AlterField(
            model_name="historicalbox",
            name="status",
            field=models.CharField(
                choices=[
                    ("open", "Open"),
                    ("verified", "Verified"),
                    ("packed", "Packed"),
                    ("shipped", "Shipped"),
                ],
                default="open",
                max_length=15,
            ),
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="contact_name",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="export_datetime",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="status",
            field=models.CharField(
                choices=[("open", "Open"), ("closed", "Closed")],
                default="open",
                max_length=15,
            ),
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="contact_name",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
        migrations.AlterField(
            model_name="manifest",
            name="export_datetime",
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="manifest",
            name="status",
            field=models.CharField(
                choices=[("open", "Open"), ("closed", "Closed")],
                default="open",
                max_length=15,
            ),
        ),
        migrations.AlterField(
            model_name="shipper",
            name="contact_name",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
    ]



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_lab_admin
from ..forms import ManifestForm
from ..models import Manifest
from .base_model_admin import BaseModelAdmin


@admin.register(Manifest, site=edc_lab_admin)
class ManifestAdmin(BaseModelAdmin, admin.ModelAdmin):

    form = ManifestForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "manifest_datetime",
                    "shipper",
                    "consignee",
                    "export_references",
                    "status",
                    "category",
                    "category_other",
                )
            },
        ),
        ("Site", {"classes": ("collapse",), "fields": ("site",)}),
        (
            "Shipping",
            {"classes": ("collapse",), "fields": ("shipped", "export_datetime")},
        ),
        audit_fieldset_tuple,
    )

    search_fields = ("manifest_identifier",)

    def get_list_filter(self, request) -> tuple:
        list_filter = super().get_list_filter(request)
        return tuple(set(("manifest_datetime",) + list_filter))

    def get_list_display(self, request) -> tuple:
        list_display = super().get_list_display(request)
        custom_fields = ("manifest_identifier", "manifest_datetime", "shipper", "consignee")
        return tuple(set(custom_fields + list_display))



from django.test import TestCase

from edc_lab.lab import (
    AliquotType,
    LabProfile,
    PanelAlreadyRegistered,
    Process,
    ProcessingProfile,
    ProcessingProfileInvalidDerivative,
    RequisitionPanel,
)

from ..models import SubjectRequisition
from .edc_lab_test_mixin import EdcLabTestMixin


class TestBuildProfile(EdcLabTestMixin, TestCase):
    def setUp(self):
        self.wb = AliquotType(name="whole_blood", numeric_code="02", alpha_code="WB")
        self.bc = AliquotType(name="buffy_coat", numeric_code="12", alpha_code="BC")

    def test_repr(self):
        obj = LabProfile(name="profile", requisition_model="edc_lab.subjectrequisition")
        self.assertTrue(repr(obj))

    def test_str(self):
        obj = LabProfile(name="profile", requisition_model="edc_lab.subjectrequisition")
        self.assertTrue(str(obj))

    def test_processing_bad(self):
        """Asserts CANNOT add process for aliquot B to a profile
        for aliquot A if B cannot be derived from A.
        """
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        b = AliquotType(name="aliquot_b", numeric_code="66", alpha_code="BB")
        process = Process(aliquot_type=b, aliquot_count=3)
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        self.assertRaises(
            ProcessingProfileInvalidDerivative,
            processing_profile.add_processes,
            process,
        )

    def test_processing_ok(self):
        """Asserts CAN add process for aliquot B to a profile
        for aliquot A since B can be derived from A.
        """
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        b = AliquotType(name="aliquot_b", numeric_code="66", alpha_code="BB")
        a.add_derivatives(b)
        process = Process(aliquot_type=b, aliquot_count=3)
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        try:
            processing_profile.add_processes(process)
        except ProcessingProfileInvalidDerivative:
            self.fail("ProcessingProfileInvalidDerivative unexpectedly raised.")

    def test_panel_adds_processing_profile(self):
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        b = AliquotType(name="aliquot_b", numeric_code="66", alpha_code="BB")
        a.add_derivatives(b)
        process = Process(aliquot_type=b, aliquot_count=3)
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        processing_profile.add_processes(process)
        panel = RequisitionPanel(name="some panel", processing_profile=processing_profile)
        lab_profile = LabProfile(
            name="profile", requisition_model="edc_lab.subjectrequisition"
        )
        lab_profile.add_panel(panel)
        self.assertEqual(panel, lab_profile.panels.get(panel.name))

    def test_add_processing(self):
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        b = AliquotType(name="aliquot_b", numeric_code="66", alpha_code="BB")
        a.add_derivatives(b)
        process = Process(aliquot_type=b, aliquot_count=3)
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        processing_profile.add_processes(process)
        panel = RequisitionPanel(name="Viral Load", processing_profile=processing_profile)
        lab_profile = LabProfile(
            name="profile", requisition_model="edc_lab.subjectrequisition"
        )
        lab_profile.add_panel(panel)

    def test_add_panel(self):
        """Assert same panel cannot be added twice."""
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        b = AliquotType(name="aliquot_b", numeric_code="66", alpha_code="BB")
        a.add_derivatives(b)
        process = Process(aliquot_type=b, aliquot_count=3)
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        processing_profile.add_processes(process)
        panel = RequisitionPanel(name="Viral Load", processing_profile=processing_profile)
        lab_profile = LabProfile(
            name="profile", requisition_model="edc_lab.subjectrequisition"
        )
        lab_profile.add_panel(panel)
        self.assertRaises(PanelAlreadyRegistered, lab_profile.add_panel, panel)

    def test_added_panel_knows_requisition_model(self):
        """Assert same panel cannot be added twice."""
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        b = AliquotType(name="aliquot_b", numeric_code="66", alpha_code="BB")
        a.add_derivatives(b)
        process = Process(aliquot_type=b, aliquot_count=3)
        processing_profile = ProcessingProfile(name="process", aliquot_type=a)
        processing_profile.add_processes(process)
        panel = RequisitionPanel(name="Viral Load", processing_profile=processing_profile)
        lab_profile = LabProfile(
            name="profile", requisition_model="edc_lab.subjectrequisition"
        )
        lab_profile.add_panel(panel)
        panel = lab_profile.panels.get("Viral Load")
        self.assertEqual(panel.requisition_model, "edc_lab.subjectrequisition")
        self.assertEqual(panel.requisition_model_cls, SubjectRequisition)

    def test_add_panel_group(self):
        panel_group = self.get_panel_group()
        lab_profile = LabProfile(
            name="profile",
            requisition_model="edc_lab.subjectrequisition",
        )
        lab_profile.add_panel(panel_group)
        panel = lab_profile.panels.get(panel_group.name)
        self.assertIsNotNone(panel)
        self.assertEqual(len(panel.utest_ids), 13)
        self.assertRaises(PanelAlreadyRegistered, lab_profile.add_panel, panel_group)

    def test_added_panel_group_knows_requisition_model(self):
        """Assert same panel cannot be added twice."""
        panel_group = self.get_panel_group()
        lab_profile = LabProfile(
            name="profile", requisition_model="edc_lab.subjectrequisition"
        )
        lab_profile.add_panel(panel_group)
        panel = lab_profile.panels.get(panel_group.name)
        self.assertEqual(panel.requisition_model, "edc_lab.subjectrequisition")
        self.assertEqual(panel.requisition_model_cls, SubjectRequisition)



from typing import Type

from django.apps import apps as django_apps
from django.db import models

from .processing_profile import ProcessingProfile


class RequisitionPanelError(Exception):
    pass


class RequisitionPanelLookupError(Exception):
    pass


class InvalidProcessingProfile(Exception):
    pass


class PanelAttrs:

    """ "A simple class of panel name attributes."""

    def __init__(self, name: str = None, alpha_code: str = None) -> None:
        title = " ".join(name.split("_")).title()
        alpha_code = alpha_code or ""
        self.abbreviation = f"{name[0:2]}{name[-1:]}".upper()
        self.verbose_name = f"{title} {alpha_code} {self.abbreviation}".replace("  ", " ")


class RequisitionPanel:

    """A panel class that contains processing profile instances."""

    panel_attrs_cls = PanelAttrs
    requisition_model: str = None  # set by lab profile.add_panel
    lab_profile_name: str = None  # set by lab profile.add_panel
    panel_model: str = "edc_lab.panel"

    def __init__(
        self,
        name: str = None,
        processing_profile: ProcessingProfile = None,
        verbose_name: str = None,
        abbreviation: str = None,
        utest_ids=None,
        is_poc=None,
        reference_range_collection_name=None,
    ) -> None:
        self._panel_model_obj = None
        self.name = name
        self.processing_profile = processing_profile
        panel_attrs = self.panel_attrs_cls(
            name=name, alpha_code=self.processing_profile.aliquot_type.alpha_code
        )
        self.abbreviation = abbreviation or panel_attrs.abbreviation
        self.verbose_name = verbose_name or panel_attrs.verbose_name
        self.utest_ids = utest_ids
        self.is_poc = is_poc
        # name for site_reportables collection, may also be set by LabProfile
        self.reference_range_collection_name = reference_range_collection_name

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, {self.aliquot_type})"

    def __str__(self):
        return self.verbose_name or self.name

    @property
    def aliquot_type(self):
        return self.processing_profile.aliquot_type

    @property
    def panel_model_cls(self):
        return django_apps.get_model(self.panel_model)

    @property
    def panel_model_obj(self):
        """Returns the underlying panel model instance."""
        if not self._panel_model_obj:
            self._panel_model_obj = self.panel_model_cls.objects.get(
                name=self.name, lab_profile_name=self.lab_profile_name
            )
        return self._panel_model_obj

    @property
    def requisition_model_cls(self) -> Type[models.Model]:
        """Returns the requisition model class associated with this
        panel (set by it's lab profile).
        """
        try:
            requisition_model_cls = django_apps.get_model(self.requisition_model)
        except (LookupError, ValueError, AttributeError) as e:
            raise RequisitionPanelLookupError(
                f"Invalid requisition model. requisition model="
                f"'{self.requisition_model}'. "
                f"See {repr(self)} or the lab profile {self.lab_profile_name}."
                f"Got {e}"
            )
        return requisition_model_cls

    @property
    def numeric_code(self) -> str:
        return self.aliquot_type.numeric_code

    @property
    def alpha_code(self) -> str:
        return self.aliquot_type.alpha_code


# TODO: panel should have some relation to the interface,
# e.g. a mapping of test_code to test_code on interface
#       for example CD4% = cd4_perc or VL = AUVL, VL = PMH



from copy import copy
from datetime import datetime

from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_appointment.models import Appointment
from edc_constants.constants import YES
from edc_facility.import_holidays import import_holidays
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED

from edc_lab import AliquotCreator, site_labs
from edc_lab.labels.aliquot_label import AliquotLabel, AliquotLabelError
from edc_lab.models import Panel
from edc_lab.tests.site_labs_test_helper import SiteLabsTestHelper

from ..models import SubjectConsent, SubjectRequisition, SubjectVisit
from ..visit_schedules import visit_schedule


class TestLabels(TestCase):

    lab_helper = SiteLabsTestHelper()

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)
        self.subject_identifier = "1111111111"
        self.gender = "M"
        self.initials = "EW"
        self.dob = datetime.now() - relativedelta(years=25)
        self.lab_helper.setup_site_labs()
        self.panel = self.lab_helper.panel

        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow(),
            identity="1111111",
            confirm_identity="1111111",
            visit_schedule_name="visit_schedule",
            schedule_name="schedule",
            dob=get_utcnow() - relativedelta(years=25),
        )
        appointment = Appointment.objects.get(visit_code="1000")
        self.subject_visit = SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )

        self.subject_requisition = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            requisition_datetime=get_utcnow(),
            drawn_datetime=get_utcnow(),
            is_drawn=YES,
            panel=Panel.objects.get(name=self.panel.name),
        )
        creator = AliquotCreator(
            subject_identifier=self.subject_identifier,
            requisition_identifier=self.subject_requisition.requisition_identifier,
            is_primary=True,
        )
        self.aliquot = creator.create(count=1, aliquot_type=self.panel.aliquot_type)

    def test_aliquot_label(self):
        label = AliquotLabel(pk=self.aliquot.pk)
        self.assertTrue(label.label_context)

    def test_aliquot_label_no_requisition_models_to_query(self):
        requisition_models = copy(site_labs.requisition_models)
        site_labs.requisition_models = []
        label = AliquotLabel(pk=self.aliquot.pk)
        try:
            label.label_context
        except AliquotLabelError:
            pass
        else:
            self.fail("AliquotLabel unexpectedly failed")
        finally:
            site_labs.requisition_models = requisition_models

    def test_aliquot_label_requisition_doesnotexist(self):
        self.aliquot.requisition_identifier = "erik"
        self.aliquot.save()
        label = AliquotLabel(pk=self.aliquot.pk)
        self.assertRaises(AliquotLabelError, getattr, label, "label_context")



from edc_auth.auth_objects import (
    ADMINISTRATION,
    AUDITOR_ROLE,
    CLINICIAN_ROLE,
    EVERYONE,
    NURSE_ROLE,
    PII_VIEW,
)
from edc_auth.site_auths import site_auths
from edc_export.auth_objects import EXPORT

from .auth_objects import (
    LAB,
    LAB_TECHNICIAN_ROLE,
    LAB_VIEW,
    lab_codenames,
    lab_view_codenames,
)

site_auths.add_group(*lab_codenames, name=LAB)
site_auths.add_group(*lab_view_codenames, name=LAB_VIEW)
site_auths.update_group(
    "edc_lab.export_aliquot",
    "edc_lab.export_box",
    "edc_lab.export_boxitem",
    "edc_lab.export_boxtype",
    "edc_lab.export_order",
    "edc_lab.export_panel",
    "edc_lab.export_result",
    "edc_lab.export_resultitem",
    name=EXPORT,
)


site_auths.add_role(ADMINISTRATION, EVERYONE, LAB, PII_VIEW, name=LAB_TECHNICIAN_ROLE)
site_auths.update_role(LAB, name=CLINICIAN_ROLE)
site_auths.update_role(LAB, name=NURSE_ROLE)
site_auths.update_role(LAB_VIEW, name=AUDITOR_ROLE)



from django.db import models
from django.db.models import PROTECT
from edc_model.validators import datetime_not_future

from edc_lab.utils import get_requisition_model_name


class CrfWithRequisitionModelMixin(models.Model):

    requisition = models.ForeignKey(
        get_requisition_model_name(),
        on_delete=PROTECT,
        related_name="+",
        verbose_name="Requisition",
        null=True,
        blank=True,
        help_text="Start typing the requisition identifier or select one from this visit",
    )

    assay_datetime = models.DateTimeField(
        verbose_name="Result Report Date and Time",
        validators=[datetime_not_future],
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True



# Generated by Django 3.2.13 on 2022-07-04 15:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_lab", "0022_auto_20211210_1839"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="historicalaliquot",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Aliquot",
                "verbose_name_plural": "historical Aliquots",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalbox",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Box",
                "verbose_name_plural": "historical Boxes",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalboxitem",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Box Item",
                "verbose_name_plural": "historical Box Items",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalconsignee",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Consignee",
                "verbose_name_plural": "historical Consignees",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalmanifest",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Manifest",
                "verbose_name_plural": "historical Manifests",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalorder",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Order",
                "verbose_name_plural": "historical Orders",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalresult",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Result",
                "verbose_name_plural": "historical Results",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalresultitem",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Result Item",
                "verbose_name_plural": "historical Result Items",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalshipper",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Shipper",
                "verbose_name_plural": "historical Shippers",
            },
        ),
        migrations.AlterField(
            model_name="historicalaliquot",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalbox",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalboxitem",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalconsignee",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalorder",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalresult",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalresultitem",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalshipper",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
    ]



# Generated by Django 2.0.1 on 2018-01-23 19:43

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("sites", "0002_alter_domain_unique"),
        ("edc_lab", "0014_auto_20180123_1936"),
    ]

    operations = [
        migrations.AddField(
            model_name="manifestitem",
            name="site",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="sites.Site",
            ),
        )
    ]



class PrimaryAliquotError(Exception):
    pass


class PrimaryAliquotPrefixError(Exception):
    pass


class PrimaryAliquot:

    """A class that gets or creates the primary aliquot."""

    def __init__(
        self,
        aliquot_type=None,
        subject_identifier=None,
        requisition_identifier=None,
        identifier_prefix=None,
        aliquot_creator_cls=None,
    ):
        self._object = None
        self.aliquot_creator_cls = aliquot_creator_cls
        self.aliquot_type = aliquot_type
        self.requisition_identifier = requisition_identifier
        self.subject_identifier = subject_identifier
        self.identifier_prefix = identifier_prefix
        self.subject_identifier = subject_identifier
        self.identifier = self.object.aliquot_identifier

    def __str__(self):
        return self.object.aliquot_identifier

    @property
    def object(self):
        """Returns an existing or newly created aliquot model instance."""
        if not self._object:
            aliquot_creator = self.aliquot_creator_cls(
                identifier_prefix=self.identifier_prefix,
                is_primary=True,
                requisition_identifier=self.requisition_identifier,
                subject_identifier=self.subject_identifier,
            )
            self._object = aliquot_creator.create_primary(aliquot_type=self.aliquot_type)
        return self._object



from django.apps import apps as django_apps
from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.urls.base import reverse
from django.utils.html import format_html

from ..constants import VERIFIED


class Manifest:
    """A class of manifest that contains boxes."""

    manifest_model = "edc_lab.manifest"
    manifest_item_model = "edc_lab.manifestitem"

    def __init__(self, manifest=None, manifest_identifier=None, request=None):
        self._manifest = manifest
        self.manifest_identifier = manifest_identifier
        self.request = request

    @property
    def manifest(self):
        if not self._manifest:
            self._manifest = self.manifest_model_cls.objects.get(
                manifest_identifier=self.manifest_identifier
            )
        return self._manifest

    @property
    def manifest_model_cls(self):
        return django_apps.get_model(self.manifest_model)

    @property
    def manifest_item_model_cls(self):
        return django_apps.get_model(self.manifest_item_model)

    def add_box(self, box=None, manifest_item_identifier=None):
        """Add a box to the manifest, if possible, and add success
        or error messages to the django message framework.
        """
        added = 0
        if not box:
            message = "Box does not exist. Got {}.".format(manifest_item_identifier)
            messages.error(self.request, message)
        elif self.validate_box_category(box) and self.validate_box_verified(box):
            try:
                manifest_item = self.manifest_item_model_cls.objects.get(
                    manifest__manifest_identifier=self.manifest.manifest_identifier,
                    identifier=manifest_item_identifier,
                )
            except ObjectDoesNotExist:
                try:
                    manifest_item = self.manifest_item_model_cls.objects.get(
                        identifier=manifest_item_identifier
                    )
                except ObjectDoesNotExist:
                    manifest_item = self.manifest_item_model_cls(
                        manifest=self.manifest, identifier=manifest_item_identifier
                    )
                    manifest_item.save()
                    added = 1
                else:
                    href = reverse(
                        self.listboard_url,
                        kwargs={
                            "manifest_identifier": manifest_item.manifest.manifest_identifier
                        },
                    )
                    manifest_identifier = manifest_item.manifest.human_readable_identifier
                    message = format_html(
                        "Item is already in a manifest. See "
                        '<a href="{}" class="alert-link">'
                        "{}</a>",
                        href,
                        manifest_identifier,
                    )
                    messages.error(self.request, message)
            else:
                message = "Duplicate item. Got {}.".format(
                    manifest_item.manifest.human_readable_identifier
                )
                messages.error(self.request, message)
        return added

    def validate_box_category(self, box=None):
        """Returns True if box category matches manifest category."""
        if box.category != self.manifest.category:
            message = "Invalid category. Manifest accepts {}. Got {}.".format(
                self.manifest.get_category_display(), box.get_category_display()
            )
            messages.error(self.request, message)
            return False
        return True

    def validate_box_verified(self, box=None):
        """Returns True if box status is verified."""
        if box.status != VERIFIED:
            message = "Box is not verified. Got {}.".format(box.human_readable_identifier)
            messages.error(self.request, message)
            return False
        return True



from django import forms
from edc_form_validators import FormValidator

from ..models import Manifest


class ManifestForm(forms.ModelForm):
    def clean(self) -> dict:
        cleaned_data = super().clean()
        form_validator = FormValidator(cleaned_data=cleaned_data)
        form_validator.validate_other_specify("category")
        return cleaned_data

    class Meta:
        fields = "__all__"
        model = Manifest



from .aliquot import Aliquot
from .box import Box, BoxIsFullError
from .box_item import BoxItem
from .box_type import BoxType
from .manifest import Consignee, Manifest, ManifestItem, Shipper
from .order import Order
from .panel import Panel
from .result import Result
from .result_item import ResultItem



from django.db import models
from edc_model import models as edc_models

from ..choices import FILL_ORDER
from ..constants import FILL_ACROSS


class BoxTypeManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, name):
        return self.get(name=name)


class BoxType(edc_models.BaseUuidModel):

    name = models.CharField(
        max_length=25, unique=True, help_text="a unique name to describe this box type"
    )

    across = models.IntegerField(
        help_text="number of cells in a row counting from left to right"
    )

    down = models.IntegerField(
        help_text="number of cells in a column counting from top to bottom"
    )

    total = models.IntegerField(help_text="total number of cells in this box type")

    fill_order = models.CharField(max_length=15, default=FILL_ACROSS, choices=FILL_ORDER)

    objects = BoxTypeManager()

    def __str__(self):
        return f"{self.name} max={self.total}"

    def natural_key(self):
        return (self.name,)  # noqa

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Box Type"
        ordering = ("name",)



from edc_model import models as edc_models
from edc_search.model_mixins import SearchSlugManager, SearchSlugModelMixin
from edc_sites.models import CurrentSiteManager

from ..managers import AliquotManager
from ..model_mixins import (
    AliquotIdentifierModelMixin,
    AliquotModelMixin,
    AliquotShippingMixin,
    AliquotTypeModelMixin,
)


class Manager(AliquotManager, SearchSlugManager):
    pass


class Aliquot(
    AliquotModelMixin,
    AliquotIdentifierModelMixin,
    AliquotTypeModelMixin,
    AliquotShippingMixin,
    SearchSlugModelMixin,
    edc_models.BaseUuidModel,
):
    def get_search_slug_fields(self):
        return [
            "aliquot_identifier",
            "human_readable_identifier",
            "subject_identifier",
            "parent_identifier",
            "requisition_identifier",
        ]

    on_site = CurrentSiteManager()

    objects = Manager()

    history = edc_models.HistoricalRecords()

    @property
    def human_readable_identifier(self):
        """Returns a human readable aliquot identifier."""
        x = self.aliquot_identifier
        return "{}-{}-{}-{}-{}".format(x[0:3], x[3:6], x[6:10], x[10:14], x[14:18])

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Aliquot"



from faker.providers import BaseProvider

from edc_lab.identifiers import RequisitionIdentifier


class EdcLabProvider(BaseProvider):
    def requisition_identifier(self):
        return RequisitionIdentifier().identifier



from django.db import models
from edc_model import models as edc_models


class ShipperManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)


class Shipper(edc_models.AddressMixin, edc_models.BaseUuidModel):

    name = models.CharField(unique=True, max_length=50)

    objects = ShipperManager()

    history = edc_models.HistoricalRecords()

    def natural_key(self):
        return (self.name,)

    def __str__(self):
        return self.name

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Shipper"
        ordering = ("name",)



from django.conf import settings
from django.test import TestCase
from edc_sites import add_or_update_django_sites
from edc_sites.single_site import SingleSite

from edc_lab.identifiers import AliquotIdentifier
from edc_lab.lab import AliquotCreator, AliquotType, PrimaryAliquot


class MyAliquotIdentifier(AliquotIdentifier):
    identifier_length = 16


class MyAliquotCreator(AliquotCreator):
    aliquot_identifier_cls = MyAliquotIdentifier


class TestPrimaryAliquot(TestCase):
    @classmethod
    def setUpClass(cls):
        add_or_update_django_sites(
            sites=[
                SingleSite(
                    settings.SITE_ID,
                    "test_site",
                    country_code="ug",
                    country="uganda",
                    domain="bugamba.ug.clinicedc.org",
                )
            ]
        )
        return super().setUpClass()

    def test_create_new_primary_aliquot(self):
        aliquot_type = AliquotType(name="aliquot_a", numeric_code="22", alpha_code="WW")
        p = PrimaryAliquot(
            requisition_identifier="ABCDE",
            identifier_prefix="066ABCDE",
            aliquot_type=aliquot_type,
            aliquot_creator_cls=MyAliquotCreator,
        )
        self.assertTrue(p.object)

    def test_get_primary_aliquot(self):
        """Asserts does not recreate aliquot model instance
        if already exists.
        """
        aliquot_type = AliquotType(name="aliquot_a", numeric_code="22", alpha_code="WW")
        p = PrimaryAliquot(
            requisition_identifier="ABCDE",
            identifier_prefix="066ABCDE",
            aliquot_type=aliquot_type,
            aliquot_creator_cls=MyAliquotCreator,
        )
        pk = p.object.id
        aliquot_identifier = p.object.aliquot_identifier
        p = PrimaryAliquot(
            requisition_identifier="ABCDE",
            identifier_prefix="066ABCDE",
            aliquot_type=aliquot_type,
            aliquot_creator_cls=MyAliquotCreator,
        )
        self.assertEqual(aliquot_identifier, p.object.aliquot_identifier)
        self.assertEqual(pk, p.object.id)

    def test_primary_aliquot_exists(self):
        """Asserts primary aliquot exists using identifier_prefix."""
        aliquot_type = AliquotType(name="aliquot_a", numeric_code="22", alpha_code="WW")
        primary_aliquot = PrimaryAliquot(
            requisition_identifier="ABCDE",
            identifier_prefix="066ABCDE",
            aliquot_type=aliquot_type,
            aliquot_creator_cls=MyAliquotCreator,
        )
        obj = primary_aliquot.object
        p = PrimaryAliquot(
            identifier_prefix=obj.identifier_prefix,
            aliquot_creator_cls=MyAliquotCreator,
        )
        self.assertEqual(obj.aliquot_identifier, p.object.aliquot_identifier)

    def test_primary_aliquot_exists2(self):
        """Asserts primary aliquot exists using requisition_identifier."""
        aliquot_type = AliquotType(name="aliquot_a", numeric_code="22", alpha_code="WW")
        primary_aliquot = PrimaryAliquot(
            requisition_identifier="ABCDE",
            identifier_prefix="066ABCDE",
            aliquot_type=aliquot_type,
            aliquot_creator_cls=MyAliquotCreator,
        )
        obj = primary_aliquot.object
        p = PrimaryAliquot(
            requisition_identifier=obj.requisition_identifier,
            aliquot_creator_cls=MyAliquotCreator,
        )
        self.assertEqual(obj.aliquot_identifier, p.object.aliquot_identifier)

    def test_str(self):
        aliquot_type = AliquotType(name="aliquot_a", numeric_code="22", alpha_code="WW")
        primary_aliquot = PrimaryAliquot(
            requisition_identifier="ABCDE",
            identifier_prefix="066ABCDE",
            aliquot_type=aliquot_type,
            aliquot_creator_cls=MyAliquotCreator,
        )
        self.assertTrue(str(primary_aliquot))



# Generated by Django 2.0.1 on 2018-01-14 12:38

import _socket
import django.db.models.deletion
import django_audit_fields.fields.uuid_auto_field
import django_revision.revision_field
import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
import edc_utils
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_lab", "0011_delete_identifierhistory")]

    operations = [
        migrations.CreateModel(
            name="Panel",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=edc_utils.date.get_utcnow),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("name", models.CharField(max_length=50)),
                ("display_name", models.CharField(max_length=50)),
                ("lab_profile_name", models.CharField(max_length=50)),
            ],
            options={"ordering": ("lab_profile_name", "name")},
        ),
        migrations.RemoveField(model_name="historicalresult", name="panel_name"),
        migrations.RemoveField(model_name="result", name="panel_name"),
        migrations.AlterUniqueTogether(
            name="panel", unique_together={("name", "lab_profile_name")}
        ),
        migrations.AddField(
            model_name="historicalresult",
            name="panel",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_lab.Panel",
            ),
        ),
        migrations.AddField(
            model_name="result",
            name="panel",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_lab.Panel",
            ),
        ),
    ]



from django.db import models
from edc_consent.field_mixins.identity_fields_mixin import IdentityFieldsMixin
from edc_consent.field_mixins.personal_fields_mixin import PersonalFieldsMixin
from edc_consent.model_mixins.consent_model_mixin import ConsentModelMixin
from edc_crf.model_mixins import CrfModelMixin
from edc_identifier.managers import SubjectIdentifierManager
from edc_identifier.model_mixins import UniqueSubjectIdentifierFieldMixin
from edc_metadata.model_mixins.creates import CreatesMetadataModelMixin
from edc_model.models import BaseUuidModel
from edc_reference import ReferenceModelConfig, site_reference_configs
from edc_registration.model_mixins import UpdatesOrCreatesRegistrationModelMixin
from edc_sites.models import SiteModelMixin
from edc_utils import get_utcnow
from edc_visit_schedule.model_mixins import (
    OffScheduleModelMixin,
    OnScheduleModelMixin,
    SubjectOnScheduleModelMixin,
    VisitScheduleFieldsModelMixin,
    VisitScheduleMethodsModelMixin,
)
from edc_visit_tracking.model_mixins import VisitModelMixin

from edc_lab.model_mixins import RequisitionModelMixin

site_reference_configs.registry = {}
reference = ReferenceModelConfig(name="edc_lab.subjectrequisition.panel", fields=["panel"])
site_reference_configs.register(reference)
reference = ReferenceModelConfig(name="edc_lab.CrfOne", fields=["f1"])
site_reference_configs.register(reference)


class SubjectRequisitionManager(models.Manager):
    def get_by_natural_key(self, requisition_identifier, subject_identifier, report_datetime):
        subject_visit = SubjectVisit.objects.get(
            subject_identifier=subject_identifier, report_datetime=report_datetime
        )
        return self.get(
            requisition_identifier=requisition_identifier, subject_visit=subject_visit
        )


class SubjectVisit(SiteModelMixin, VisitModelMixin, CreatesMetadataModelMixin, BaseUuidModel):
    def update_reference_on_save(self):
        pass


class SubjectRequisition(RequisitionModelMixin, BaseUuidModel):
    def update_reference_on_save(self):
        pass

    class Meta(RequisitionModelMixin.Meta):
        pass


class OnSchedule(SiteModelMixin, OnScheduleModelMixin, BaseUuidModel):

    pass


class OffSchedule(SiteModelMixin, OffScheduleModelMixin, BaseUuidModel):

    pass


class DeathReport(SiteModelMixin, UniqueSubjectIdentifierFieldMixin, BaseUuidModel):

    objects = SubjectIdentifierManager()

    def natural_key(self):
        return (self.subject_identifier,)


class SubjectConsent(
    SiteModelMixin,
    ConsentModelMixin,
    UniqueSubjectIdentifierFieldMixin,
    PersonalFieldsMixin,
    IdentityFieldsMixin,
    UpdatesOrCreatesRegistrationModelMixin,
    SubjectOnScheduleModelMixin,
    VisitScheduleFieldsModelMixin,
    VisitScheduleMethodsModelMixin,
    BaseUuidModel,
):

    objects = SubjectIdentifierManager()

    def natural_key(self):
        return (self.subject_identifier,)

    class Meta(ConsentModelMixin.Meta):
        pass


class CrfOne(CrfModelMixin, BaseUuidModel):

    dte = models.DateTimeField(default=get_utcnow)



from django.test import TestCase

from edc_lab.identifiers import (
    AliquotIdentifier,
    AliquotIdentifierCountError,
    AliquotIdentifierLengthError,
    Prefix,
    PrefixKeyError,
    PrefixLengthError,
)


class TestAliquotPrefix(TestCase):
    def test_prefix(self):
        prefix_obj = Prefix(template="{opt1}{opt2}", length=8, opt1="opt1", opt2="opt2")
        self.assertEqual(str(prefix_obj), "opt1opt2")

    def test_prefix_invalid_length(self):
        self.assertRaises(
            PrefixLengthError,
            Prefix,
            template="{opt1}{opt2}",
            length=7,
            opt1="opt1",
            opt2="opt2",
        )

    def test_prefix_missing_opt(self):
        self.assertRaises(
            PrefixKeyError, Prefix, template="{opt1}{opt2}", length=8, opt1="opt1"
        )


class TestAliquotIdentifier(TestCase):
    def test_valid_length(self):
        class MyAliquotIdentifier(AliquotIdentifier):
            identifier_length = 0

        self.assertRaises(
            AliquotIdentifierLengthError,
            MyAliquotIdentifier,
            identifier_prefix="1234567890",
            numeric_code="22",
        )

    def test_length_raises(self):
        """Asserts raises exception for invalid identifier length."""

        class MyAliquotIdentifier(AliquotIdentifier):
            identifier_length = 0

        self.assertRaises(
            AliquotIdentifierLengthError,
            MyAliquotIdentifier,
            identifier_prefix="1234567890",
        )

    def test_numeric_code(self):
        class MyAliquotIdentifier(AliquotIdentifier):
            identifier_length = 16

        identifier = MyAliquotIdentifier(identifier_prefix="XXXXXXXX", numeric_code="02")
        self.assertIn("02", str(identifier))

    def test_primary(self):
        class MyAliquotIdentifier(AliquotIdentifier):
            identifier_length = 16

        identifier = MyAliquotIdentifier(identifier_prefix="XXXXXXXX", numeric_code="11")
        self.assertIn("0000", str(identifier))
        self.assertTrue(identifier.is_primary)

    def test_not_primary_needs_count(self):
        """Asserts need a count if not a primary aliquot."""

        class MyAliquotIdentifier(AliquotIdentifier):
            identifier_length = 16

        self.assertRaises(
            AliquotIdentifierCountError,
            MyAliquotIdentifier,
            parent_segment="0201",
            identifier_prefix="XXXXXXXX",
            numeric_code="11",
        )

    def test_not_primary_parent_segment(self):
        class MyAliquotIdentifier(AliquotIdentifier):
            identifier_length = 16

        identifier = MyAliquotIdentifier(
            parent_segment="0201",
            identifier_prefix="XXXXXXXX",
            numeric_code="11",
            count=2,
        )
        self.assertIn("0201", str(identifier))
        self.assertFalse(identifier.is_primary)

    def test_not_primary_parent_segment2(self):
        class MyAliquotIdentifier(AliquotIdentifier):
            identifier_length = 16

        identifier = MyAliquotIdentifier(
            parent_segment="0201",
            identifier_prefix="XXXXXXXX",
            numeric_code="11",
            count=2,
        )
        self.assertIn("1102", str(identifier))
        self.assertFalse(identifier.is_primary)

    def test_large_count_raises_length_error(self):
        class MyAliquotIdentifier(AliquotIdentifier):
            identifier_length = 16

        self.assertRaises(
            AliquotIdentifierLengthError,
            MyAliquotIdentifier,
            parent_segment="0201",
            identifier_prefix="XXXXXXXX",
            numeric_code="11",
            count=222,
        )

    def test_large_count_valid(self):
        class MyAliquotIdentifier(AliquotIdentifier):
            identifier_length = 17

        try:
            MyAliquotIdentifier(
                parent_segment="0201",
                identifier_prefix="XXXXXXXX",
                numeric_code="11",
                count=222,
            )
        except AliquotIdentifierLengthError:
            self.fail("AliquotIdentifierLengthError unexpectedly raised.")



from django.db import models
from edc_model import models as edc_models


class PanelManager(models.Manager):

    use_in_migrations = True

    def get_by_natural_key(self, name, lab_profile_name):
        return self.get(name=name, lab_profile_name=lab_profile_name)


class Panel(edc_models.BaseUuidModel):

    name = models.CharField(max_length=50)

    display_name = models.CharField(max_length=50)

    lab_profile_name = models.CharField(max_length=50)

    objects = PanelManager()

    def __str__(self):
        return self.display_name or self.name

    def natural_key(self):
        return (self.name, self.lab_profile_name)

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Panel"
        unique_together = ("name", "lab_profile_name")
        ordering = ("lab_profile_name", "name")



from django.conf import settings
from django.contrib.sites.models import Site
from django.test import TestCase
from edc_appointment.models import Appointment
from edc_constants.constants import NO, YES
from edc_facility import import_holidays
from edc_sites import add_or_update_django_sites
from edc_sites.single_site import SingleSite
from edc_sites.tests import SiteTestCaseMixin
from edc_utils.date import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_visit_tracking.constants import SCHEDULED

from edc_lab.identifiers import AliquotIdentifier as AliquotIdentifierBase
from edc_lab.lab import AliquotCreator as AliquotCreatorBase
from edc_lab.lab import AliquotType, Process, ProcessingProfile
from edc_lab.lab import Specimen as SpecimenBase
from edc_lab.lab import SpecimenNotDrawnError, SpecimenProcessor
from edc_lab.models import Aliquot

from ..models import SubjectConsent, SubjectRequisition, SubjectVisit
from ..site_labs_test_helper import SiteLabsTestHelper
from ..visit_schedules import visit_schedule


class AliquotIdentifier(AliquotIdentifierBase):
    identifier_length = 18


class AliquotCreator(AliquotCreatorBase):
    aliquot_identifier_cls = AliquotIdentifier


class Specimen(SpecimenBase):
    aliquot_creator_cls = AliquotCreator


class TestSpecimen(SiteTestCaseMixin, TestCase):

    lab_helper = SiteLabsTestHelper()

    @classmethod
    def setUpClass(cls):
        add_or_update_django_sites(
            sites=[
                SingleSite(
                    settings.SITE_ID,
                    "test_site",
                    country_code="ug",
                    country="uganda",
                    domain="bugamba.ug.clinicedc.org",
                )
            ]
        )
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)
        self.subject_identifier = "1111111111"
        self.lab_helper.setup_site_labs()
        self.panel = self.lab_helper.panel
        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow(),
            identity="1111111",
            confirm_identity="1111111",
            visit_schedule_name="visit_schedule",
            schedule_name="schedule",
            site=Site.objects.get_current(),
        )
        appointment = Appointment.objects.get(visit_code="1000")
        self.subject_visit = SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )

    def test_specimen_processor(self):
        SpecimenProcessor(aliquot_creator_cls=AliquotCreator)

    def test_specimen(self):
        requisition = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel.panel_model_obj,
            protocol_number="999",
            is_drawn=YES,
        )
        Specimen(requisition=requisition)

    def test_specimen_repr(self):
        requisition = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel.panel_model_obj,
            protocol_number="999",
            is_drawn=YES,
        )
        specimen = Specimen(requisition=requisition)
        self.assertTrue(repr(specimen))

    def test_specimen_from_pk(self):
        requisition = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel.panel_model_obj,
            protocol_number="999",
            is_drawn=YES,
        )
        Specimen(requisition=requisition)

    def test_specimen_not_drawn(self):
        requisition = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel.panel_model_obj,
            protocol_number="999",
            is_drawn=NO,
        )
        self.assertRaises(SpecimenNotDrawnError, Specimen, requisition=requisition)


class TestSpecimen2(SiteTestCaseMixin, TestCase):

    lab_helper = SiteLabsTestHelper()

    @classmethod
    def setUpClass(cls):
        add_or_update_django_sites(
            sites=[
                SingleSite(
                    settings.SITE_ID,
                    "test_site",
                    country_code="ug",
                    country="uganda",
                    domain="bugamba.ug.clinicedc.org",
                )
            ]
        )
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        self.lab_helper.setup_site_labs()
        self.panel = self.lab_helper.panel
        self.profile_aliquot_count = self.lab_helper.profile_aliquot_count
        self.subject_identifier = "1111111111"
        Site.objects.get_current()
        SubjectConsent.objects.create(
            subject_identifier=self.subject_identifier,
            consent_datetime=get_utcnow(),
            identity="1111111",
            confirm_identity="1111111",
            visit_schedule_name="visit_schedule",
            schedule_name="schedule",
        )
        appointment = Appointment.objects.get(visit_code="1000")
        self.subject_visit = SubjectVisit.objects.create(
            appointment=appointment, report_datetime=get_utcnow(), reason=SCHEDULED
        )
        self.requisition = SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel.panel_model_obj,
            protocol_number="999",
            is_drawn=YES,
        )
        self.specimen = Specimen(requisition=self.requisition)

    def test_requisition_creates_aliquot(self):
        """Asserts passing requisition to specimen class
        creates an aliquot.
        """
        requisition = SubjectRequisition.objects.get(pk=self.requisition.pk)
        self.assertEqual(
            Aliquot.objects.filter(
                requisition_identifier=requisition.requisition_identifier,
                is_primary=True,
            ).count(),
            1,
        )

    def test_requisition_gets_aliquot(self):
        """Asserts passing requisition to specimen class gets
        an existing aliquot.
        """
        # instantiate again, to get primary aliquot
        specimen = Specimen(requisition=self.requisition)
        obj = Aliquot.objects.get(
            requisition_identifier=self.requisition.requisition_identifier,
            is_primary=True,
        )
        self.assertEqual(specimen.aliquots[0].aliquot_identifier, obj.aliquot_identifier)

    def test_process_repr(self):
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        process = Process(aliquot_type=a)
        self.assertTrue(repr(process))

    def test_process_profile_repr(self):
        a = AliquotType(name="aliquot_a", numeric_code="55", alpha_code="AA")
        processing_profile = ProcessingProfile(name="processing_profile", aliquot_type=a)
        self.assertTrue(repr(processing_profile))

    def test_specimen_process(self):
        """Asserts calling process creates the correct number
        of child aliquots.
        """
        self.assertEqual(self.specimen.aliquots.count(), 1)
        self.specimen.process()
        self.assertEqual(self.specimen.aliquots.count(), self.profile_aliquot_count + 1)

    def test_specimen_process2(self):
        """Asserts calling process more than once has no effect."""
        self.specimen.process()
        self.assertEqual(self.specimen.aliquots.count(), self.profile_aliquot_count + 1)
        self.specimen.process()
        self.specimen.process()
        self.assertEqual(self.specimen.aliquots.count(), self.profile_aliquot_count + 1)

    def test_specimen_process_identifier_prefix(self):
        """Assert all aliquots start with the correct identifier
        prefix.
        """
        self.specimen.process()
        for aliquot in self.specimen.aliquots.order_by("created"):
            self.assertIn(
                self.specimen.primary_aliquot.identifier_prefix,
                aliquot.aliquot_identifier,
            )

    def test_specimen_process_identifier_parent_segment(self):
        """Assert all aliquots have correct 4 chars parent_segment."""
        self.specimen.process()
        parent_segment = self.specimen.primary_aliquot.aliquot_identifier[-4:]

        aliquot = self.specimen.aliquots.order_by("count")[0]
        self.assertTrue(aliquot.is_primary)
        self.assertEqual("0000", aliquot.aliquot_identifier[-8:-4])

        for aliquot in self.specimen.aliquots.order_by("count")[1:]:
            self.assertFalse(aliquot.is_primary)
            self.assertEqual(parent_segment, aliquot.aliquot_identifier[-8:-4])

    def test_specimen_process_identifier_child_segment(self):
        """Assert all aliquots have correct 4 chars child_segment."""
        self.specimen.process()

        aliquot = self.specimen.aliquots.order_by("count")[0]
        self.assertTrue(aliquot.is_primary)
        self.assertEqual("5501", aliquot.aliquot_identifier[-4:])

        for index, aliquot in enumerate(self.specimen.aliquots.order_by("count")[1:]):
            index += 2
            self.assertFalse(aliquot.is_primary)
            self.assertEqual(f"66{str(index).zfill(2)}", aliquot.aliquot_identifier[-4:])



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-05 17:40
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_lab", "0002_auto_20170305_1939")]

    operations = [
        migrations.AddField(
            model_name="historicalmanifest",
            name="consignee_name",
            field=models.CharField(
                blank=True,
                max_length=50,
                null=True,
                verbose_name="Consignee's contact name",
            ),
        ),
        migrations.AddField(
            model_name="historicalmanifest",
            name="shipper_name",
            field=models.CharField(
                blank=True,
                max_length=50,
                null=True,
                verbose_name="Shipper's contact name",
            ),
        ),
        migrations.AddField(
            model_name="manifest",
            name="consignee_name",
            field=models.CharField(
                blank=True,
                max_length=50,
                null=True,
                verbose_name="Consignee's contact name",
            ),
        ),
        migrations.AddField(
            model_name="manifest",
            name="shipper_name",
            field=models.CharField(
                blank=True,
                max_length=50,
                null=True,
                verbose_name="Shipper's contact name",
            ),
        ),
    ]



from django.db import models
from django.utils import timezone
from edc_sites.models import SiteModelMixin


class ResultItemModelMixin(SiteModelMixin, models.Model):

    report_datetime = models.DateTimeField(null=True)

    utestid = models.CharField(max_length=25, null=True)

    value = models.CharField(max_length=25, null=True)

    quantifier = models.CharField(max_length=25, null=True)

    value_datetime = models.DateTimeField(null=True)

    reference = models.CharField(max_length=25, null=True)

    pending_datetime = models.DateTimeField(default=timezone.now)

    pending = models.BooleanField(default=True)

    resulted_datetime = models.DateTimeField(null=True)

    resulted = models.BooleanField(default=False)

    class Meta:
        abstract = True



from django.apps import apps as django_apps
from django.conf import settings


def get_requisition_model():
    return django_apps.get_model(get_requisition_model_name())


def get_panel_model():
    return django_apps.get_model(get_panel_model_name())


def get_panel_model_name():
    return getattr(settings, "LAB_PANEL_MODEL", "edc_lab.panel")


def get_requisition_model_name():
    return settings.SUBJECT_REQUISITION_MODEL


def get_subject_requisition_model():
    return get_requisition_model_name()



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple

from ..admin_site import edc_lab_admin
from ..forms import BoxForm
from ..models import Box
from .base_model_admin import BaseModelAdmin


@admin.register(Box, site=edc_lab_admin)
class BoxAdmin(BaseModelAdmin, admin.ModelAdmin):

    form = BoxForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "name",
                    "box_type",
                    "specimen_types",
                    "box_datetime",
                    "category",
                    "category_other",
                    "accept_primary",
                    "comment",
                )
            },
        ),
        (
            "Status",
            {
                "classes": ("collapse",),
                "fields": ("status", "verified", "verified_datetime"),
            },
        ),
        audit_fieldset_tuple,
    )

    radio_fields = {
        "box_type": admin.VERTICAL,
        "category": admin.VERTICAL,
        "status": admin.VERTICAL,
    }

    def get_list_display(self, request) -> tuple:
        list_display = super().get_list_display(request)
        custom_fields = (
            "box_identifier",
            "name",
            "category",
            "specimen_types",
            "box_type",
            "box_datetime",
            "user_created",
            "user_modified",
        )
        return tuple(set(custom_fields + list_display))

    def get_list_filter(self, request) -> tuple:
        list_filter = super().get_list_filter(request)
        custom_fields = ("box_datetime", "specimen_types", "category", "box_type")
        return tuple(set(custom_fields + list_filter))

    def get_readonly_fields(self, request, obj=None) -> tuple:
        readonly_fields = super().get_readonly_fields(request, obj=obj)
        return tuple(set(readonly_fields + ("status", "verified", "verified_datetime")))



from django.db import models
from django.db.models.deletion import PROTECT
from edc_model import models as edc_models
from edc_sites.models import CurrentSiteManager

from ..model_mixins import ResultItemModelMixin
from .result import Result


class ResultItemManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, report_datetime, requisition_identifier):
        return self.get(
            report_datetime=report_datetime,
            requisition__requisition_identifier=requisition_identifier,
        )


class ResultItem(ResultItemModelMixin, edc_models.BaseUuidModel):

    result = models.ForeignKey(Result, on_delete=PROTECT)

    on_site = CurrentSiteManager()

    objects = ResultItemManager()

    history = edc_models.HistoricalRecords()

    def natural_key(self):
        return (self.report_datetime,) + self.result.natural_key()

    natural_key.dependencies = ["edc_lab.result", "sites.Site"]

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Result Item"



CONDITION_OK = "OK"
FILL_ACROSS = "across"
FILL_DOWN = "down"
FINGER_PRICK = "finger_prick"
PACKED = "packed"
SHIPPED = "shipped"
STORAGE = "storage"
TESTING = "testing"
TUBE = "tube"
VERIFIED = "verified"
WHOLE_BLOOD = "02"
EQ = "="
GT = ">"
GTE = ">="
LT = "<"
LTE = "<="



# Generated by Django 3.2.9 on 2021-12-10 15:39

from django.db import migrations
import edc_lab.models.aliquot
import edc_lab.models.box
import edc_lab.models.box_item
import edc_lab.models.box_type
import edc_lab.models.manifest.manifest
import edc_lab.models.order
import edc_lab.models.panel
import edc_lab.models.result
import edc_lab.models.result_item
import edc_sites.models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_lab", "0021_auto_20200513_0034"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="aliquot",
            managers=[
                ("on_site", edc_sites.models.CurrentSiteManager()),
                ("objects", edc_lab.models.aliquot.Manager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="box",
            managers=[
                ("on_site", edc_sites.models.CurrentSiteManager()),
                ("objects", edc_lab.models.box.BoxManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="boxitem",
            managers=[
                ("objects", edc_lab.models.box_item.BoxItemManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="boxtype",
            managers=[
                ("objects", edc_lab.models.box_type.BoxTypeManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="manifest",
            managers=[
                ("on_site", edc_sites.models.CurrentSiteManager()),
                ("objects", edc_lab.models.manifest.manifest.Manager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="order",
            managers=[
                ("on_site", edc_sites.models.CurrentSiteManager()),
                ("objects", edc_lab.models.order.OrderManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="panel",
            managers=[
                ("objects", edc_lab.models.panel.PanelManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="result",
            managers=[
                ("on_site", edc_sites.models.CurrentSiteManager()),
                ("objects", edc_lab.models.result.ResultManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="resultitem",
            managers=[
                ("on_site", edc_sites.models.CurrentSiteManager()),
                ("objects", edc_lab.models.result_item.ResultItemManager()),
            ],
        ),
    ]



from django.db import models
from django.db.models.deletion import PROTECT
from edc_model import models as edc_models
from edc_search.model_mixins import SearchSlugManager, SearchSlugModelMixin
from edc_sites.models import SiteModelMixin

from ...model_mixins import VerifyModelMixin
from .manifest import Manifest


class ManifestItemManager(SearchSlugManager, models.Manager):
    def get_by_natural_key(self, identifier, manifest_identifier):
        return self.get(identifier=identifier, manifest_identifier=manifest_identifier)


class ManifestItem(
    SiteModelMixin, SearchSlugModelMixin, VerifyModelMixin, edc_models.BaseUuidModel
):
    def get_search_slug_fields(self):
        return ["identifier", "human_readable_identifier"]

    manifest = models.ForeignKey(Manifest, on_delete=PROTECT)

    identifier = models.CharField(max_length=25)

    comment = models.CharField(max_length=25, null=True, blank=True)

    objects = ManifestItemManager()

    def natural_key(self):
        return (self.identifier,) + self.manifest.natural_key()

    natural_key.dependencies = ["edc_lab.manifest", "sites.Site"]

    @property
    def human_readable_identifier(self):
        x = self.identifier
        return "{}-{}-{}".format(x[0:4], x[4:8], x[8:12])

    class Meta(edc_models.BaseUuidModel.Meta):
        verbose_name = "Manifest Item"
        ordering = ("created",)
        unique_together = ("manifest", "identifier")



from django.conf import settings
from django.test import TestCase
from edc_sites import add_or_update_django_sites
from edc_sites.single_site import SingleSite
from edc_sites.tests import SiteTestCaseMixin
from edc_test_utils.natural_key_test_helper import NaturalKeyTestHelper


class TestNaturalKey(SiteTestCaseMixin, TestCase):

    nk_test_helper = NaturalKeyTestHelper()

    @classmethod
    def setUpClass(cls):
        add_or_update_django_sites(
            sites=[
                SingleSite(
                    settings.SITE_ID,
                    "test_site",
                    country_code="ug",
                    country="uganda",
                    domain="bugamba.ug.clinicedc.org",
                )
            ]
        )
        return super().setUpClass()

    def test_natural_key_attrs(self):
        self.nk_test_helper.nk_test_natural_key_attr("edc_lab")

    def test_get_by_natural_key_attr(self):
        self.nk_test_helper.nk_test_get_by_natural_key_attr("edc_lab")



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-05 18:15
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("edc_lab", "0003_auto_20170305_1940")]

    operations = [
        migrations.RemoveField(model_name="historicalmanifest", name="consignee_name"),
        migrations.RemoveField(model_name="historicalmanifest", name="shipper_name"),
        migrations.RemoveField(model_name="manifest", name="consignee_name"),
        migrations.RemoveField(model_name="manifest", name="shipper_name"),
        migrations.AddField(
            model_name="consignee",
            name="contact_name",
            field=models.CharField(default="1", max_length=50),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="historicalconsignee",
            name="contact_name",
            field=models.CharField(default="1", max_length=50),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="historicalshipper",
            name="contact_name",
            field=models.CharField(
                blank=True,
                help_text="If left blank will use user's first and last name and the time the manifest is created.",
                max_length=50,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="shipper",
            name="contact_name",
            field=models.CharField(
                blank=True,
                help_text="If left blank will use user's first and last name and the time the manifest is created.",
                max_length=50,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="description",
            field=models.TextField(
                help_text="If blank will be automatically generated",
                null=True,
                verbose_name="Description of contents",
            ),
        ),
        migrations.AlterField(
            model_name="historicalmanifest",
            name="export_references",
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="manifest",
            name="description",
            field=models.TextField(
                help_text="If blank will be automatically generated",
                null=True,
                verbose_name="Description of contents",
            ),
        ),
        migrations.AlterField(
            model_name="manifest",
            name="export_references",
            field=models.TextField(blank=True, null=True),
        ),
    ]
